package corefdb

import (
	"errors"
	"io"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/lazywriter"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl/filepack"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/hashmdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"

	"github.com/tidwall/sjson"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/utiliymdl/guidmdl"

	"github.com/tidwall/buntdb"
	"github.com/tidwall/gjson"
)

const (
	lineBreak = "\r\n"

	// BucketTypeSimple -
	BucketTypeSimple = "Normal"
	// BucketTypePack -
	BucketTypePack = "Pack"
	// BucketTypeAppend -
	BucketTypeAppend = `Append`
	// BucketTypeMixed -
	BucketTypeMixed = `Mixed`
	// BucketTypeAsset -
	BucketTypeAsset = `Asset`
	// IndexKeyValSeperator -
	IndexKeyValSeperator = "="
	// FileType - represents key for type of file. Used whenever we need to set file type field in json
	FileType = "fileType"
	// FileTypeAsset - represents media file type
	FileTypeAsset = "Asset"
	// FileTypeFDBIndex -
	FileTypeFDBIndex = "FDBIndex"

	MigrationTypeUpdate     = "MigrationTypeUpdate"
	MigrationTypeReplace    = "MigrationTypeReplace"
	MigrationTypeKeyword    = "migrationType"
	MigrationConfigFilename = "migrationConfig"
)

// ErrNoDataFound - This error describes that the required data might be deleted and not found. Kindly ignore this error in caller.
var ErrNoDataFound = errors.New("No data found")
var bucketLock = sync.Mutex{}

// Bucket - Bucket
type Bucket struct {
	BucketID        string                          `json:"bucketId"`
	IsDynamicName   bool                            `json:"isDynamicName"`
	BucketNameQuery string                          `json:"bucketNameQuery"`
	Indices         []string                        `json:"indices"`
	BucketPath      string                          `json:"bucketPath"`
	InFileIndexMap  map[string]filepack.InFileIndex `json:"inFileIndexMap"`
	BucketType      string                          `json:"bucketType"`
	EnableLazy      bool                            `json:"enableLazy"`
}

// Index - Index
type Index struct {
	indexObj       *buntdb.DB
	IndexID        string       `json:"indexId"`
	IndexNameQuery string       `json:"indexNameQuery"`
	BucketSequence []string     `json:"bucketSequence"`
	IndexFields    []IndexField `json:"indexFields"`
	IsDynamicName  bool         `json:"isDynamicName"`
}

// IndexField - IndexField
type IndexField struct {
	FieldName string `json:"fieldName"`
	Query     string `json:"query"`
}

// GetNewIndex - GetNewIndex returns new index
func (fdb *FDB) GetNewIndex(indexNameQuery string, IsDynamicName bool) (*Index, error) {
	fdb.indexMux.Lock()
	defer fdb.indexMux.Unlock()
	if _, ok := fdb.indices[indexNameQuery]; ok {
		return nil, errormdl.Wrap("Index name already found")
	}
	idx := &Index{
		IndexNameQuery: indexNameQuery,
		IndexID:        indexNameQuery,
		IsDynamicName:  IsDynamicName,
	}
	db, err := fdb.GetIndexDB(idx)
	if err != nil {
		return nil, err
	}
	idx.indexObj = db

	fdb.indices[idx.IndexID] = idx
	return idx, nil
}

// RegisterNewIndex - RegisterNewIndex returns new index
func (fdb *FDB) RegisterNewIndex(idx *Index) (*Index, error) {
	fdb.indexMux.Lock()
	defer fdb.indexMux.Unlock()
	if _, ok := fdb.indices[idx.IndexID]; ok {
		return nil, errormdl.Wrap("Index ID already found")
	}

	db, err := fdb.GetIndexDB(idx)
	if err != nil {
		return nil, err
	}
	idx.indexObj = db

	fdb.indices[idx.IndexID] = idx
	return idx, nil
}

// GetIndexDB -  GetIndexDB
func (fdb *FDB) GetIndexDB(index *Index) (*buntdb.DB, error) {
	// dbPath := filepath.Join(fdb.DBPath, INDEXFOLDER)
	// filemdl.CreateDirectoryRecursive(dbPath)
	// dbPath = filepath.Join(dbPath, index.IndexID)
	db, err := buntdb.Open(":memory:")
	if err != nil {
		return nil, err
	}
	return db, nil
}

// GetNewBucket - return Bucket Obj with provided data and new guid
func (fdb *FDB) GetNewBucket(BucketNameQuery string, IsDynamicName bool, parentBucket *Bucket) *Bucket {
	bucket := &Bucket{
		BucketID:        guidmdl.GetGUID(),
		BucketNameQuery: BucketNameQuery,
		IsDynamicName:   IsDynamicName,
	}
	if IsDynamicName {
		BucketNameQuery = "$$" + BucketNameQuery
	}
	bucket.BucketPath = filepath.Join(parentBucket.BucketPath, BucketNameQuery)
	fdb.buckets[bucket.BucketID] = bucket
	return bucket
}

// NewBucket - return Bucket Obj
func (fdb *FDB) NewBucket(bucket, parentBucket *Bucket) *Bucket {
	fdb.bLocker.Lock()
	defer fdb.bLocker.Unlock()

	bucketNameQuery := bucket.BucketNameQuery
	if bucket.IsDynamicName {
		bucketNameQuery = "$$" + bucket.BucketNameQuery
	}
	bucket.BucketPath = filepath.Join(parentBucket.BucketPath, bucketNameQuery)
	fdb.buckets[bucket.BucketID] = bucket
	return bucket
}

// // SetIsBucketPackType - SetIsBucketPackType
// func (bucket *Bucket) SetIsBucketPackType(isPackType bool) {
// 	bucket.IsSinglePackType = isPackType
// }

// SetBucketType -
func (bucket *Bucket) SetBucketType(bucketType string) {
	bucket.BucketType = bucketType
}

// EnableLazyWrite - enables lazy on the bucket
func (bucket *Bucket) EnableLazyWrite(isEnable bool) {
	bucket.EnableLazy = isEnable
}

// SetInFileIndex - SetInFileIndex
func (bucket *Bucket) SetInFileIndex(inFileIndex filepack.InFileIndex) {
	if bucket.InFileIndexMap == nil {
		bucket.InFileIndexMap = make(map[string]filepack.InFileIndex)
	}
	bucket.InFileIndexMap[inFileIndex.FileType] = inFileIndex
}

// SetInFileIndexes - SetInFileIndexes
func (bucket *Bucket) SetInFileIndexes(inFileIndexes []filepack.InFileIndex) {
	if bucket.InFileIndexMap == nil {
		bucket.InFileIndexMap = make(map[string]filepack.InFileIndex, len(inFileIndexes))
	}
	for _, inFileIndex := range inFileIndexes {
		bucket.InFileIndexMap[inFileIndex.FileType] = inFileIndex
	}
}

// SetBucket - set bucket in index
func (index *Index) SetBucket(bucket *Bucket) *Index {
	if bucket != nil {
		index.BucketSequence = append(index.BucketSequence, bucket.BucketID)
		bucket.Indices = append(bucket.Indices, index.IndexID)
	}
	return index
}

// SetFields - SetFields
func (index *Index) SetFields(indexField ...IndexField) *Index {
	index.IndexFields = indexField
	return index
}

// bucketNameResolver - returns bucket name
func (fdb *FDB) bucketNameResolver(bucketID string, rs *gjson.Result) (string, error) {
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return "", errormdl.Wrap("Bucket not found: " + bucketID)
	}
	name := bucket.BucketNameQuery
	if bucket.IsDynamicName {
		name = rs.Get(name).String()
	}
	if name == "" {
		return name, errormdl.Wrap("Bucket name should not be empty: " + bucket.BucketNameQuery)
	}
	return name, nil
}

// resolveIndex- returns bucket path by resolving index
func (fdb *FDB) resolveIndex(index *Index, rs *gjson.Result) (string, error) {
	path := ""
	for _, bucketID := range index.BucketSequence {
		bucketName, err := fdb.bucketNameResolver(bucketID, rs)
		if err != nil {
			return path, err
		}
		path = filepath.Join(path, bucketName)
	}
	indexName := index.IndexNameQuery
	if index.IsDynamicName {
		indexName = rs.Get(index.IndexNameQuery).String()
	}
	if indexName == "" {
		return path, errormdl.Wrap("required attribute not provided:" + index.IndexNameQuery)
	}
	path = filepath.Join(path, indexName)
	return path, nil
}

// createIndexJSON - create JSON with index field data
func createIndexJSON(index *Index, rs *gjson.Result) (string, error) {
	json := `{}`
	for _, indexField := range index.IndexFields {
		if rs.Get(indexField.Query).Value() == nil {
			return json, errormdl.Wrap("please provide value for index field: " + indexField.Query)
		}
		json, _ = sjson.Set(json, indexField.FieldName, rs.Get(indexField.Query).Value())
	}
	return json, nil
}

// updateIndexJSON - update JSON with index field data
func updateIndexJSON(index *Index, existingData string, rs *gjson.Result) (string, error) {
	json := existingData
	var err error
	for _, indexField := range index.IndexFields {
		if rs.Get(indexField.Query).Value() == nil {
			continue
		}
		json, err = sjson.Set(json, indexField.FieldName, rs.Get(indexField.Query).Value())
		// loggermdl.LogError("Error - ", err)
	}
	return json, err
}

// UpdateIndexLazyObjectInCache - updates index data in lay writer cache
func UpdateIndexLazyObjectInCache(indexID string, indexData *Index) error {
	// lazy cache must be present for provided indexID
	lazyObj, ok := IndexLazyObjHolder.Get(indexID)
	if !ok {
		loggermdl.LogError("index not found in lazy writer cache")
		return errormdl.Wrap("index not found in lazy writer cache")
	}

	idxLazyData, ok := lazyObj.(lazywriter.LazyCacheObject)
	if !ok {
		loggermdl.LogError("interface type is not lazywriter.LazyCacheObject")
		return errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
	}

	// idxLazyData.GJSONData = indexData
	idxLazyData.InterfaceData = indexData
	if ok := IndexMaster.SaveOrUpdateDataInCache(idxLazyData); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return errormdl.Wrap("failed to update index data in lazy cache")
	}

	IndexLazyObjHolder.SetNoExpiration(indexID, idxLazyData)
	return nil
}

// AppendDataInLazyObjectInCache - updates index data in lay writer cache
func AppendDataInLazyObjectInCache(bucketID string, data gjson.Result, filePath string) error {
	// lazy cache must be present for provided bucketID
	bucketLock.Lock()
	defer bucketLock.Unlock()
	lazyObj, ok := AppendLazyObjHolder.Get(bucketID)
	if !ok {
		loggermdl.LogError(AppendLazyObjHolder.GetItems())
		return errormdl.Wrap("lazyObject not found in lazywriter cache")
	}

	bucketLazyData, ok := lazyObj.(lazywriter.LazyCacheObject)
	if !ok {
		loggermdl.LogError("interface type is not lazywriter.LazyCacheObject")
		return errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
	}

	bucketLazyData.FileName = filePath
	updatedData := ""
	var err error
	dataBytes := []byte(data.String())
	if isSecurityEnabled {
		_, fileName := filepath.Split(filePath)
		dataBytes, err = encryptData(dataBytes, fileName)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
	}
	if bucketLazyData.InterfaceData == nil {
		updatedData = string(dataBytes)
	} else {
		previousData, ok := bucketLazyData.InterfaceData.(string)
		if !ok {
			return errormdl.Wrap("previous lazy data is not of type string")
		}
		updatedData = previousData + string(lineBreak) + string(dataBytes)
	}

	// bucketLazyData.GJSONData = indexData
	bucketLazyData.InterfaceData = updatedData
	if ok := AppendMaster.SaveOrUpdateDataInCache(bucketLazyData); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return errormdl.Wrap("failed to update index data in lazy cache")
	}

	AppendLazyObjHolder.SetNoExpiration(bucketID, bucketLazyData)
	return nil
}

// // restoreFilesFromFilePack - restoreFilesFromFilePack
// func restoreFilesFromFilePack(fileName string, fileData []byte) error {
// 	if len(fileData) > 0 {
// 		err := filemdl.WriteFile(fileName, fileData, true, false)
// 		if errormdl.CheckErr(err) != nil {
// 			loggermdl.LogError(err)
// 			return errormdl.CheckErr(err)
// 		}
// 	}
// 	return nil
// }

// func readNormalFile(path string) ([]byte, error) {
// 	ba, err := filemdl.ReadFile(path)
// 	if err != nil {
// 		loggermdl.LogError(err)
// 		return nil, err
// 	}
// 	if isSecurityEnabled {
// 		_, fileName := filepath.Split(path)
// 		ba, err = decryptData(ba, fileName)
// 		if err != nil {
// 			loggermdl.LogError(err)
// 			return nil, err
// 		}
// 	}
// 	return ba, nil
// }

func readNormalFileUsingFp(fp *os.File, rs *gjson.Result, secParams securitymdl.FDBSecParams) ([]byte, error) {
	ba, err := filemdl.ReadFileUsingFp(fp)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}

	if len(ba) == 0 {
		return ba, nil
	}

	if secParams.EnableSecurity {
		// _, fileName := filepath.Split(fp.Name())
		key, err := securitymdl.GenSecKeyBytes(fp.Name(), rs)
		if err != nil {
			loggermdl.LogError("failed to generate security key: ", err)
			return nil, err
		}
		ba, err = decryptwithDecompression(ba, secParams.EnableCompression, key)
		if err != nil {
			loggermdl.LogError(err)
			return nil, err
		}
	}
	return ba, nil
}

// // readNormalFiles - reads file and returns file data
// func (fdb *FDB) readNormalFiles(path []string) (*gjson.Result, error) {
// 	sb := strings.Builder{}
// 	sb.WriteString("[")
// 	for i := 0; i < len(path); i++ {
// 		updatedPath := filepath.Join(fdb.DBPath, path[i])
// 		ba := []byte{'{', '}'}
// 		var err error
// 		if filemdl.FileAvailabilityCheck(updatedPath) {
// 			ba, err = readNormalFile(updatedPath)
// 			if err != nil {
// 				loggermdl.LogError(err)
// 				return nil, err
// 			}
// 			// loggermdl.LogError("File data for index = ", i, " - ", string(ba))
// 			_, err = sb.Write(ba)
// 			if err != nil {
// 				loggermdl.LogError(err)
// 				return nil, err
// 			}
// 			sb.WriteString(",")
// 		}
// 	}
// 	sb.WriteString("]")
// 	finalResult := strings.Replace(sb.String(), ",]", "]", 1)
// 	rs := gjson.Parse(finalResult)
// 	return &rs, nil
// }

// readNormalFilesUsingFp - reads file and returns file data

func readNormalFilesUsingFp(fps []*os.File, bucket *Bucket, rs *gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, error) {

	sb := strings.Builder{}
	sb.WriteString("[")
	for _, fp := range fps {
		if filemdl.FileAvailabilityCheck(fp.Name()) {
			if bucket.EnableLazy {
				data, err := readDataFromLazyBucket(bucket, fp, rs, secParams)
				if err != nil {
					loggermdl.LogError(err)
					return nil, err
				}
				_, err = sb.WriteString(data)
				if err != nil {
					loggermdl.LogError(err)
					return nil, err
				}
			} else {
				ba, err := readNormalFileUsingFp(fp, rs, secParams)
				if err != nil {
					loggermdl.LogError(err)
					return nil, err
				}
				// loggermdl.LogError("File data for index = ", i, " - ", string(ba))
				_, err = sb.Write(ba)
				if err != nil {
					loggermdl.LogError(err)
					return nil, err
				}
			}
			sb.WriteString(",")
		}
	}
	sb.WriteString("]")
	finalResult := strings.Replace(sb.String(), ",]", "]", 1)
	res := gjson.Parse(finalResult)
	return &res, nil
}

// // GetDataByConcat - GetDataByConcat
// func GetDataByConcat(dbName, indexID string, query ...string) (*gjson.Result, error) {
// 	fdb, err := GetFDBInstance(dbName)
// 	if err != nil {
// 		return nil, err
// 	}
// 	index, ok := fdb.GetFDBIndex(indexID)
// 	if !ok {
// 		loggermdl.LogError("INDEX not found: " + indexID)
// 		return nil, errormdl.Wrap("INDEX not found: " + indexID)
// 	}
// 	// path := []string{}
// 	sb := strings.Builder{}
// 	sb.WriteString("[")
// 	index.indexObj.View(func(tx *buntdb.Tx) error {

// 		return tx.Ascend(indexID, func(key, value string) bool {
// 			sb.WriteString((value + ","))
// 			return true
// 		})

// 		// return nil
// 	})
// 	sb.WriteString("]")
// 	finalData := strings.Replace(sb.String(), ",]", "]", 1)
// 	rsJSON := gjson.Parse(finalData)
// 	for i := 0; i < len(query); i++ {
// 		rsJSON = rsJSON.Get(query[i] + "#")
// 		if rsJSON.Value() == nil {
// 			break
// 		}
// 	}

// 	return nil, nil
// }
func prepareLazyDataForNormalBucket(prevData string, data *gjson.Result) (string, error) {
	return data.String(), nil
}

type lazySec struct {
	securitymdl.FDBSecParams
}

func (ls lazySec) SaveFnForNormalBucket(bucketId string, data *lazywriter.LazyCacheObject) {
	cachedData, ok := data.InterfaceData.(string)
	if !ok {
		loggermdl.LogError("lazy write: no data found in lazy cache for bucket id ", bucketId)
		return
	}

	var (
		cachedDataRS = gjson.Parse(cachedData)
		bucketType   = cachedDataRS.Get("bucketType").String()
		dataRS       = cachedDataRS.Get("data")
		dataBytes    []byte
		err          error
	)

	if bucketType != BucketTypeSimple {
		loggermdl.LogError("lazy write: expected Normal bucket, got ", bucketType)
		return
	}

	dataBytes = []byte(dataRS.String())

	if ls.EnableSecurity {
		key, err := securitymdl.GenSecKeyBytes(data.FileName, &dataRS)
		if err != nil {
			loggermdl.LogError("lazy write: failed to generate security key for normal bucket, ", err)
			return
		}

		dataBytes, err = encryptWithCompression(dataBytes, ls.EnableCompression, key)
		if err != nil {
			loggermdl.LogError("lazy write: failed to encrypt data for normal bucket, ", err)
			return
		}
	}

	if err = filemdl.WriteFile(data.FileName, dataBytes, true, true); err != nil {
		loggermdl.LogError("lazy write: failed to save data from normal bucket to disk, ", err)
	}
}

var lazyCallBackFn lazywriter.SaveDataFn = func(bucketId string, data *lazywriter.LazyCacheObject) {
	dataInLazyMemory, ok := data.InterfaceData.(string)
	if !ok {
		loggermdl.LogError("data not found to write")
		return
	}
	dataObj := gjson.Parse(dataInLazyMemory)
	bucketType := dataObj.Get("bucketType").String()
	dataByte := []byte(dataObj.Get("data").String())
	var err error
	switch bucketType {
	// case BucketTypeSimple:
	// 	err = filemdl.WriteFile(data.FileName, dataByte, true, true)
	case BucketTypeAppend:
		_, _, err = filemdl.AppendDataToFile(data.FileName, dataByte, true)
		if err != nil {
			loggermdl.LogError(err)
		}
		err = LazyWriteObjMaster.ClearLazyObjInterfaceData(data.FileName)
		data.InterfaceData = ""
		LazyWriteObjHolder.Set(data.FileName, data)

	default:
		loggermdl.LogError("operation not allowed on this bucket type: ", bucketType)
	}
	if err != nil {
		loggermdl.LogError(err)
	}
}

func saveDatInLazyBucket(bucket *Bucket, fp *os.File, data *gjson.Result, secParams securitymdl.FDBSecParams) error {
	if !(bucket.BucketType == BucketTypeSimple || bucket.BucketType == BucketTypeAppend) {
		return errormdl.Wrap("save to lazy cache operation not allowed on this bucket type: " + bucket.BucketType)
	}

	var (
		lazyDataObj lazywriter.LazyCacheObject
		prevDataStr = ""
	)

	lazyObj, ok := LazyWriteObjHolder.Get(fp.Name())
	if ok {
		lazyDataObj, ok = lazyObj.(lazywriter.LazyCacheObject)
		if !ok {
			return errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
		}
		prevDataStr, ok = lazyDataObj.InterfaceData.(string)
		if !ok {
			return errormdl.Wrap("previous lazy data is not of type string")
		}
	} else {
		lazyDataObj = lazywriter.LazyCacheObject{
			FileName:      fp.Name(),
			Identifier:    fp.Name(),
			InterfaceData: "",
			SaveFn: func() lazywriter.SaveDataFn {
				if bucket.BucketType == BucketTypeAppend {
					return lazyCallBackFn
				}

				return (lazySec{secParams}).SaveFnForNormalBucket
			}(),
		}
	}

	updatedData := ""

	// prevDataObj.Get("data")
	switch bucket.BucketType {
	case BucketTypeSimple:
		updatedData = data.String()
	case BucketTypeAppend:
		prevDataObj := gjson.Parse(prevDataStr)
		prevData := prevDataObj.Get("data").String()
		dataToSave := data.String()
		// // currently the security is disabled for append bucket. Don't have any way to read encrypted data if encrypted by field query.
		// rs := gjson.Parse(dataToSave)
		// if secParams.EnableSecurity {
		// 	// _, fileName := filepath.Split(f.Name())
		// 	key, err := securitymdl.GenSecKeyBytes(fp.Name(), &rs)
		// 	if err != nil {
		// 		loggermdl.LogError("failed to generate security key for append bucket: ", err)
		// 		return err
		// 	}

		// 	dataBytes, err := encryptWithCompression([]byte(dataToSave), secParams.EnableCompression, key)
		// 	if err != nil {
		// 		loggermdl.LogError("failed to encrypt data for append bucket: ", err)
		// 		return err
		// 	}

		// 	dataToSave = string(dataBytes)
		// }
		updatedData = prevData + dataToSave + string(lineBreak)
	default:
		return errormdl.Wrap("operation not allowed on this bucket type: " + bucket.BucketType)
	}

	updatedDataObj, _ := sjson.Set("", "bucketType", bucket.BucketType)
	updatedDataObj, _ = sjson.Set(updatedDataObj, "data", updatedData)

	// bucketLazyData.GJSONData = indexData
	lazyDataObj.InterfaceData = updatedDataObj
	if ok := LazyWriteObjMaster.SaveOrUpdateDataInCache(lazyDataObj); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return errormdl.Wrap("failed to update index data in lazy cache")
	}

	LazyWriteObjHolder.SetNoExpiration(fp.Name(), lazyDataObj)
	return nil
}

func updateDataInLazyBucket(bucket *Bucket, fp *os.File, data *gjson.Result, secParam securitymdl.FDBSecParams) (*gjson.Result, error) {
	if bucket.BucketType != BucketTypeSimple {
		return data, errormdl.Wrap("update data in lazy cache operation not allowed on this bucket type: " + bucket.BucketType)
	}
	var lazyDataObj lazywriter.LazyCacheObject
	lazyObj, foundLazyObj := LazyWriteObjHolder.Get(fp.Name())
	updatedData := ""
	prevDataStr := ""
	ok := false
	if foundLazyObj {
		lazyDataObj, ok = lazyObj.(lazywriter.LazyCacheObject)
		if !ok {
			return nil, errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
		}
		prevDataStr, ok = lazyDataObj.InterfaceData.(string)
		if !ok {
			return nil, errormdl.Wrap("previous lazy data is not of type string")
		}
		updatedData = gjson.Parse(prevDataStr).Get("data").String()
	} else {
		lazyDataObj = lazywriter.LazyCacheObject{
			FileName:   fp.Name(),
			Identifier: fp.Name(),
			SaveFn:     (lazySec{secParam}).SaveFnForNormalBucket,
		}
		dataByte, err := readNormalFileUsingFp(fp, data, secParam)

		if err != nil {
			return data, err
		}
		updatedData = string(dataByte)
	}

	data.ForEach(func(key, val gjson.Result) bool {
		updatedData, _ = sjson.Set(updatedData, key.String(), val.Value())
		return true
	})

	updatedDataStr, _ := sjson.Set("", "bucketType", bucket.BucketType)
	updatedDataStr, _ = sjson.Set(updatedDataStr, "data", updatedData)

	// bucketLazyData.GJSONData = indexData
	lazyDataObj.InterfaceData = updatedDataStr
	if ok := LazyWriteObjMaster.SaveOrUpdateDataInCache(lazyDataObj); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return nil, errormdl.Wrap("failed to update index data in lazy cache")
	}

	LazyWriteObjHolder.SetNoExpiration(fp.Name(), lazyDataObj)
	updatedDataObj := gjson.Parse(updatedData)
	return &updatedDataObj, nil
}

func readDataFromLazyBucket(bucket *Bucket, fp *os.File, rs *gjson.Result, secParam securitymdl.FDBSecParams) (data string, err error) {
	if bucket.BucketType != BucketTypeSimple {
		return data, errormdl.Wrap("read from lazy cache operation not allowed on this bucket type: " + bucket.BucketType)
	}
	var lazyDataObj lazywriter.LazyCacheObject
	lazyObj, foundLazyObj := LazyWriteObjHolder.Get(fp.Name())
	prevDataStr := ""
	ok := false
	if foundLazyObj {
		lazyDataObj, ok = lazyObj.(lazywriter.LazyCacheObject)
		if !ok {
			return data, errormdl.Wrap("interface type is not lazywriter.LazyCacheObject")
		}
		prevDataStr, ok = lazyDataObj.InterfaceData.(string)
		if !ok {
			return data, errormdl.Wrap("previous lazy data is not of type string")
		}
		prevData := gjson.Parse(prevDataStr)
		data = prevData.Get("data").String()

		return
	}
	dataByte, err := readNormalFileUsingFp(fp, rs, secParam)
	if err != nil {
		return data, err
	}
	data = string(dataByte)
	lazyDataObj = lazywriter.LazyCacheObject{
		FileName:   fp.Name(),
		Identifier: fp.Name(),
		SaveFn:     (lazySec{secParam}).SaveFnForNormalBucket,
	}

	updatedDataObj, _ := sjson.Set("", "bucketType", bucket.BucketType)
	updatedDataObj, _ = sjson.Set(updatedDataObj, "data", data)

	// bucketLazyData.GJSONData = indexData
	lazyDataObj.InterfaceData = updatedDataObj
	if ok := LazyWriteObjMaster.SaveOrUpdateDataInCache(lazyDataObj); !ok {
		loggermdl.LogError("failed to update index data in lazy cache")
		return data, errormdl.Wrap("failed to update index data in lazy cache")
	}
	LazyWriteObjHolder.SetNoExpiration(fp.Name(), lazyDataObj)

	return data, nil
}

func deleteDataFromLazyBucket(bucket *Bucket, fp *os.File) error {
	if bucket.BucketType != BucketTypeSimple {
		return errormdl.Wrap("operation not allowed on this bucket type: " + bucket.BucketType)
	}

	LazyWriteObjMaster.ClearLazyObjInterfaceData(fp.Name())
	LazyWriteObjHolder.Delete(fp.Name())

	return nil
}

// func saveDataInNormalBucketUsingFpNew(f *os.File, bucket *Bucket, rs *gjson.Result) error {
// 	dataBytes := []byte(rs.String())
// 	var err error
// 	if bucket.EnableLazy {
// 		return saveDatInLazyBucket(bucket, f, rs)
// 	}
// 	if isSecurityEnabled {
// 		_, fileName := filepath.Split(f.Name())
// 		dataBytes, err = encryptData(dataBytes, fileName)
// 		if err != nil {
// 			loggermdl.LogError(err)
// 			return err
// 		}
// 	}
// 	err = filemdl.WriteFileUsingFp(f, dataBytes, true, false)
// 	if errormdl.CheckErr(err) != nil {
// 		loggermdl.LogError(err)
// 		return errormdl.CheckErr(err)
// 	}
// 	return nil
// }

func saveDataInNormalBucketUsingFp(f *os.File, bucket *Bucket, rs *gjson.Result, secParams securitymdl.FDBSecParams) error {
	dataBytes := []byte(rs.String())
	var err error
	if bucket.EnableLazy {
		if secParams.EnableSecurity {
			// do not add data to lazy cache if security is enabled and got empty result for field query
			if _, err := securitymdl.CheckDataForFieldQuery(f.Name(), rs); err != nil {
				loggermdl.LogError("failed to save data in lazy writer chache: ", err)
				return err
			}
		}

		return saveDatInLazyBucket(bucket, f, rs, secParams)
	}
	if secParams.EnableSecurity {
		// _, fileName := filepath.Split(f.Name())
		key, err := securitymdl.GenSecKeyBytes(f.Name(), rs)
		if err != nil {
			loggermdl.LogError("failed to generate security key for normal bucket: ", err)
			return err
		}

		dataBytes, err = encryptWithCompression(dataBytes, secParams.EnableCompression, key)
		if err != nil {
			loggermdl.LogError("failed to encrypt data for normal bucket: ", err)
			return err
		}
	}
	err = filemdl.WriteFileUsingFp(f, dataBytes, true, false)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	return nil
}

// func saveDataInNormalBucket(filePath string, rs *gjson.Result) error {
// 	dataBytes := []byte(rs.String())
// 	var err error
// 	if isSecurityEnabled {
// 		_, fileName := filepath.Split(filePath)
// 		dataBytes, err = encryptData(dataBytes, fileName)
// 		if err != nil {
// 			loggermdl.LogError(err)
// 			return err
// 		}
// 	}
// 	err = filemdl.WriteFile(filePath, dataBytes, true, false)
// 	if errormdl.CheckErr(err) != nil {
// 		loggermdl.LogError(err)
// 		return errormdl.CheckErr(err)
// 	}
// 	return nil
// }

// func saveDataInPackBucket(fdb *FDB, filePath string, index *Index, rs *gjson.Result) error {
// 	// check is file locked
// 	// if yes
// 	// wait
// 	// else
// 	// /*  */update filestatus to fileStatusUpdatingData
// 	// compress and encrypt data
// 	// calculate filehash
// 	//save data in file & getoffset and len
// 	//update footer address
// 	// update filestatuse to fileStatusUpdatingIndex
// 	// update index
// 	// update filestatus to fileStatusReady

// 	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 	bucket := fdb.buckets[bucketID]
// 	fileType := rs.Get("fileType").String()
// 	if len(fileType) == 0 {
// 		return errormdl.Wrap("please specify fileType")
// 	}

// 	isFilePresent := filemdl.FileAvailabilityCheck(filePath)

// 	infileIndex, ok := bucket.InFileIndexMap[fileType]
// 	if !ok {
// 		loggermdl.LogError("infileIndex for specified fileType not found: ", fileType)
// 		return errormdl.Wrap("infileIndex for specified fileType not found: " + fileType)
// 	}

// 	err := filepack.AddFileInPackFile(filePath, infileIndex.IndexFields, rs)
// 	if err != nil {
// 		loggermdl.LogError("fail to add fdb index entry in file: ", err)
// 		return errormdl.Wrap("fail to write data in file")
// 	}
// 	// for adding fdb index data
// 	if !isFilePresent {
// 		err = addFDBIndexEntryFile(filePath, bucket, index.IndexFields, rs)
// 	}
// 	return err
// }

func saveDataInPackBucketUsingFp(fdb *FDB, f *os.File, index *Index, infileIndexData, rs *gjson.Result) (*gjson.Result, error) {
	// check is file locked
	// if yes
	// wait
	// else
	// /*  */update filestatus to fileStatusUpdatingData
	// compress and encrypt data
	// calculate filehash
	//save data in file & getoffset and len
	//update footer address
	// update filestatuse to fileStatusUpdatingIndex
	// update index
	// update filestatus to fileStatusReady
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket := fdb.buckets[bucketID]
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		return infileIndexData, errormdl.Wrap("please specify fileType")
	}
	isFilePresent := filemdl.FileAvailabilityCheck(f.Name())
	info, err := f.Stat()
	if err != nil {
		loggermdl.LogError(err)
		return infileIndexData, errormdl.Wrap("failed to save data: " + err.Error())
	}

	infileIndex, ok := bucket.InFileIndexMap[fileType]
	if !ok {
		loggermdl.LogError("infileIndex for specified fileType not found: ", fileType)
		return infileIndexData, errormdl.Wrap("infileIndex for specified fileType not found: " + fileType)
	}

	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}

	infileIndexData, err = filepack.AddPackFileUsingFp(f, infileIndex.IndexFields, infileIndexData, rs, secParams)
	if err != nil {
		loggermdl.LogError("fail to add fdb index entry in file: ", err)
		return infileIndexData, errormdl.Wrap("fail to write data in file")
	}
	// for adding fdb index data
	if !isFilePresent || info.Size() == 0 {
		err = addFDBIndexEntryFileUsingFp(f, bucket, index.IndexFields, rs, secParams)
		if err != nil {
			loggermdl.LogError("fail to add fdb index entry in file: ", err)
			return infileIndexData, errormdl.Wrap("fail to write data in file")
		}
	}
	return infileIndexData, err
}

// func saveDataInPackBucketUsingFpOld(fdb *FDB, f *os.File, index *Index, rs *gjson.Result) error {
// 	// check is file locked
// 	// if yes
// 	// wait
// 	// else
// 	// /*  */update filestatus to fileStatusUpdatingData
// 	// compress and encrypt data
// 	// calculate filehash
// 	//save data in file & getoffset and len
// 	//update footer address
// 	// update filestatuse to fileStatusUpdatingIndex
// 	// update index
// 	// update filestatus to fileStatusReady

// 	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 	bucket := fdb.buckets[bucketID]
// 	fileType := rs.Get("fileType").String()
// 	if len(fileType) == 0 {
// 		return errormdl.Wrap("please specify fileType")
// 	}
// 	isFilePresent := filemdl.FileAvailabilityCheck(f.Name())
// 	info, err := f.Stat()
// 	if err != nil {
// 		loggermdl.LogError(err)
// 		return errormdl.Wrap("failed to save data: " + err.Error())
// 	}

// 	infileIndex, ok := bucket.InFileIndexMap[fileType]
// 	if !ok {
// 		loggermdl.LogError("infileIndex for specified fileType not found: ", fileType)
// 		return errormdl.Wrap("infileIndex for specified fileType not found: " + fileType)
// 	}

// 	err = filepack.AddFileInPackFileUsingFp(f, infileIndex.IndexFields, rs)
// 	if err != nil {
// 		loggermdl.LogError("fail to add fdb index entry in file: ", err)
// 		return errormdl.Wrap("fail to write data in file")
// 	}
// 	// for adding fdb index data
// 	if !isFilePresent || info.Size() == 0 {
// 		err = addFDBIndexEntryFileUsingFp(f, bucket, index.IndexFields, rs)
// 	}
// 	return err
// }

// SaveMediaInFDB -
func SaveMediaInFDB(dbName string, indexID string, mediaData []byte, rs *gjson.Result) (string, error) {
	recordPath := ""
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return recordPath, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return recordPath, errormdl.Wrap("INDEX not found: " + indexID)
	}
	path, err := fdb.resolveIndex(index, rs)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return recordPath, errormdl.CheckErr(err)
	}
	rsStr, _ := sjson.Set(rs.String(), "fileType", FileTypeAsset)
	data := gjson.Parse(rsStr)

	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return recordPath, errormdl.Wrap("Bucket not found: " + bucketID)
	}

	if bucket.BucketType != BucketTypeMixed && bucket.BucketType != BucketTypeAsset {
		loggermdl.LogError("operation not valid on bucket type: ", bucket.BucketType)
		return recordPath, errormdl.Wrap("operation not valid on this type of bucket: " + bucket.BucketType)
	}
	prevFDBIndexVal, err := GetFDBIndexEntryByPath(index, path)
	if err != nil {
		loggermdl.LogError("fail to get fdb index entry ", bucket.BucketType)
		return recordPath, errormdl.Wrap("fail to add media " + bucket.BucketType)
	}

	filePath := filepath.Join(fdb.DBPath, path)
	var f *os.File
	var inFileIndexData *gjson.Result
	rfile, err := fileFpCache.Get(filePath)
	if err != nil {
		f, err = openFile(filePath)
		if err != nil {
			loggermdl.LogError(err)
			return recordPath, err
		}
		rfile = fileFpCache.Set(f, nil)
	} else {
		f = rfile.file
		inFileIndexData = rfile.InfileIndex
	}
	rfile.lock.Lock()
	defer rfile.lock.Unlock()
	recordID, updatedInfileIndex, err := filepack.AddMediaInPackFileUsingFp(f, inFileIndexData, mediaData, []filepack.InFileIndexField{}, &data)
	if err != nil {
		loggermdl.LogError("fail to add fdb index entry in file: ", err)
		return recordPath, errormdl.Wrap("fail to add fdb index entry")
	}
	fileFpCache.Set(f, updatedInfileIndex)
	// json, err := createIndexJSON(index, rs)
	// if err != nil {
	// 	loggermdl.LogError(err)
	// 	return recordPath, err
	// }
	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	// if index.IsDynamicName {
	// 	rowID = rs.Get(index.IndexNameQuery).String()
	// } else {
	// 	rowID = index.IndexNameQuery
	// }
	prevFDBIndexVal, _ = sjson.Set(prevFDBIndexVal, "rowID", rowID)
	updatedJSON, err := updateIndexJSON(index, prevFDBIndexVal, rs)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	updatedJSONObj := gjson.Parse(updatedJSON)
	err = updateFDBIndexEntry(index, path, &updatedJSONObj)
	// _, _, err = tx.Set(path, updatedJSON, nil)
	if err != nil {
		loggermdl.LogError("failed to update index data - ", err)
		return "", errormdl.Wrap("failed to update index data")
	}

	if isLazyWriterEnabled {
		err = UpdateIndexLazyObjectInCache(indexID, index)
		if err != nil {
			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
			return "", errormdl.Wrap("failed to update index data in lazy writer cache")
		}
	}

	if err != nil {
		loggermdl.LogError("fail to set fdb index: ", err)
		return recordPath, errormdl.Wrap("fail to set fdb index: " + err.Error())
	}
	recordPath = dbName + "/" + indexID + "/" + rowID + "/" + recordID
	return recordPath, nil
}

// UpdateMediaInFDB -
func UpdateMediaInFDB(dbName string, indexID, recordID string, mediaData []byte, rs *gjson.Result) (string, error) {
	recordID = strings.TrimSpace(recordID)
	if recordID == "" {
		return "", errors.New("please provide recordID for Update operation")
	}
	recordPath := ""
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return recordPath, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return recordPath, errormdl.Wrap("INDEX not found: " + indexID)
	}
	path, err := fdb.resolveIndex(index, rs)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return recordPath, errormdl.CheckErr(err)
	}
	rsStr, _ := sjson.Set(rs.String(), "fileType", FileTypeAsset)
	data := gjson.Parse(rsStr)

	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return recordPath, errormdl.Wrap("Bucket not found: " + bucketID)
	}

	if bucket.BucketType != BucketTypeMixed && bucket.BucketType != BucketTypeAsset {
		loggermdl.LogError("operation not valid on bucket type: ", bucket.BucketType)
		return recordPath, errormdl.Wrap("operation not valid on this type of bucket: " + bucket.BucketType)
	}

	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	// rowID := ""
	// if index.IsDynamicName {
	// 	rowID = rs.Get(index.IndexNameQuery).String()
	// } else {
	// 	rowID = index.IndexNameQuery
	// }

	if rowID == "" {
		loggermdl.LogError("please provide value for index field :", index.IndexNameQuery)
		return recordPath, errormdl.Wrap("please provide value for index field :" + index.IndexNameQuery)
	}
	query := []string{`#[rowID=` + rowID + `]`}
	// rs := gjson.Result{}
	filePath := ""
	prevIndexVal := ""
	// loggermdl.LogDebug(query)
	err = index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(query); i++ {
				rsJSON = rsJSON.Get(query[i] + "#")
			}
			if rsJSON.Value() != nil && rsJSON.IsArray() && len(rsJSON.Array()) > 0 {
				filePath = key
				prevIndexVal = value
			}
			return true
		})
	})

	if err != nil {
		loggermdl.LogError("failed to get data: ", err)
		return recordPath, errormdl.Wrap("data not found: " + err.Error())
	}

	if filePath == "" {
		loggermdl.LogError("no data found to update")
		return recordPath, errormdl.Wrap("no data found to update")
	}

	filePath = filepath.Join(fdb.DBPath, filePath)

	var f *os.File
	var inFileIndexData *gjson.Result
	rfile, err := fileFpCache.Get(filePath)
	if err != nil {
		f, err = openFile(filePath)
		if err != nil {
			loggermdl.LogError(err)
			return recordPath, err
		}
		rfile = fileFpCache.Set(f, nil)
	} else {
		f = rfile.file
		inFileIndexData = rfile.InfileIndex
	}
	rfile.lock.Lock()
	defer rfile.lock.Unlock()
	var updatedInfileIndex *gjson.Result
	recordID, updatedInfileIndex, err = filepack.UpdateMediaInPackFileUsingFp(f, inFileIndexData, recordID, mediaData, []filepack.InFileIndexField{}, &data)
	if err != nil {
		loggermdl.LogError("fail to update media: ", err)
		return recordPath, err
	}
	fileFpCache.Set(f, updatedInfileIndex)

	updatedJSON, err := updateIndexJSON(index, prevIndexVal, rs)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	updatedJSONObj := gjson.Parse(updatedJSON)
	err = updateFDBIndexEntry(index, path, &updatedJSONObj)
	// _, _, err = tx.Set(path, updatedJSON, nil)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}

	if isLazyWriterEnabled {
		err = UpdateIndexLazyObjectInCache(indexID, index)
		if err != nil {
			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
			return "", errormdl.Wrap("failed to update index data in lazy writer cache")
		}
	}

	if err != nil {
		loggermdl.LogError("fail to set fdb index: ", err)
		return recordPath, errormdl.Wrap("fail to set fdb index: " + err.Error())
	}
	recordPath = dbName + "/" + indexID + "/" + rowID + "/" + recordID
	return recordPath, nil
}

// UpsertMediaInFDB -
func UpsertMediaInFDB(dbName string, indexID, recordID string, mediaData []byte, rs *gjson.Result) (string, error) {
	recordID = strings.TrimSpace(recordID)
	if recordID == "" {
		return "", errors.New("please provide recordID for Upsert operation")
	}
	recordPath := ""
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return recordPath, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return recordPath, errormdl.Wrap("INDEX not found: " + indexID)
	}
	path, err := fdb.resolveIndex(index, rs)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return recordPath, errormdl.CheckErr(err)
	}
	rsStr, _ := sjson.Set(rs.String(), "fileType", FileTypeAsset)
	data := gjson.Parse(rsStr)

	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return recordPath, errormdl.Wrap("Bucket not found: " + bucketID)
	}

	if bucket.BucketType != BucketTypeMixed && bucket.BucketType != BucketTypeAsset {
		loggermdl.LogError("operation not valid on bucket type: ", bucket.BucketType)
		return recordPath, errormdl.Wrap("operation not valid on this type of bucket: " + bucket.BucketType)
	}

	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}

	// rowID := ""
	// if index.IsDynamicName {
	// 	rowID = rs.Get(index.IndexNameQuery).String()
	// } else {
	// 	rowID = index.IndexNameQuery
	// }
	query := []string{`#[rowID=` + rowID + `]`}
	// rs := gjson.Result{}
	prevIndexVal := ""

	// loggermdl.LogDebug(query)
	err = index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(query); i++ {
				rsJSON = rsJSON.Get(query[i] + "#")
			}
			if rsJSON.Value() != nil && rsJSON.IsArray() && len(rsJSON.Array()) > 0 {
				prevIndexVal = value
			}
			return true
		})
	})

	if err != nil {
		loggermdl.LogError("fail to upsert media: ", err)
		return recordPath, errormdl.Wrap("fail to upsert media: " + err.Error())
	}

	filePath := filepath.Join(fdb.DBPath, path)
	var f *os.File
	var inFileIndexData *gjson.Result
	rfile, err := fileFpCache.Get(filePath)
	if err != nil {
		f, err = openFile(filePath)
		if err != nil {
			loggermdl.LogError(err)
			return recordPath, err
		}
		rfile = fileFpCache.Set(f, nil)
	} else {
		f = rfile.file
		inFileIndexData = rfile.InfileIndex
	}

	rfile.lock.Lock()
	defer rfile.lock.Unlock()
	var updatedInfileIndex *gjson.Result

	recordID, updatedInfileIndex, err = filepack.UpsertMediaInPackFileUsingFp(f, inFileIndexData, recordID, mediaData, []filepack.InFileIndexField{}, &data)
	// if filePathInFdbIndex == "" {
	// 	recordID, updatedInfileIndex, err = filepack.AddMediaInPackFileUsingFp(f, inFileIndexData, mediaData, []filepack.InFileIndexField{}, &data)
	// } else {
	// 	recordID, updatedInfileIndex, err = filepack.UpdateMediaInPackFileUsingFp(f, inFileIndexData, recordID, mediaData, []filepack.InFileIndexField{}, &data)
	// }

	// recordID, updatedInfileIndex, err = filepack.UpsertMediaInPackFileUsingFp(f, inFileIndexData, recordID, mediaData, []filepack.InFileIndexField{}, &data)
	if err != nil {
		loggermdl.LogError("fail to upsert media: ", err)
		return recordPath, errormdl.Wrap("fail to upsert media: " + err.Error())
	}
	fileFpCache.Set(f, updatedInfileIndex)
	prevIndexVal, _ = sjson.Set(prevIndexVal, "rowID", rowID)

	updatedJSON, err := updateIndexJSON(index, prevIndexVal, rs)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}
	updatedJSONObj := gjson.Parse(updatedJSON)
	err = updateFDBIndexEntry(index, path, &updatedJSONObj)
	// _, _, err = tx.Set(path, updatedJSON, nil)
	if err != nil {
		loggermdl.LogError(err)
		return recordPath, err
	}

	if err != nil {
		loggermdl.LogError("failed to upsert index data - ", err)
		return "", errormdl.Wrap("failed to upsert index data")
	}

	if isLazyWriterEnabled {
		err = UpdateIndexLazyObjectInCache(indexID, index)
		if err != nil {
			loggermdl.LogError("failed to upsert media: ", err)
			return "", errormdl.Wrap("failed to upsert media: fail to update lazy writer cache")
		}
	}

	if err != nil {
		loggermdl.LogError("failed to upsert media: fail to set fdb index: ", err)
		return recordPath, errormdl.Wrap("fail to set fdb index: " + err.Error())
	}

	recordPath = dbName + "/" + indexID + "/" + rowID + "/" + recordID
	return recordPath, nil
}

// GetMediaFromFDB -
func GetMediaFromFDB(dbName, indexID, rowID, recordID string) ([]byte, gjson.Result, error) {
	dataByte := []byte{}
	fileMeta := gjson.Result{}
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return dataByte, fileMeta, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return dataByte, fileMeta, errormdl.Wrap("INDEX not found: " + indexID)
	}
	query := []string{`#[rowID=` + rowID + `]`}
	// rs := gjson.Result{}

	filePaths, err := getFDBIndexEntry(index, query)

	if err != nil {
		loggermdl.LogError("failed to get data: ", err)
		return dataByte, fileMeta, errormdl.Wrap("data not found: " + err.Error())
	}

	if len(filePaths) == 0 {
		loggermdl.LogError("data not found")
		return dataByte, fileMeta, errormdl.Wrap("data not found")
	}
	// bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	// bucket, ok := fdb.buckets[bucketID]
	// if !ok {
	// 	loggermdl.LogError("Bucket not found: " + bucketID)
	// 	return dataByte,fileMeta errormdl.Wrap("Bucket not found: " + bucketID)
	// }
	var f *os.File
	var fpInfileIndex *gjson.Result
	filePath := filepath.Join(fdb.DBPath, filePaths[0])
	rfile, err := fileFpCache.Get(filePath)
	if err != nil {
		f, err = openFile(filePath)
		if err != nil {
			return dataByte, fileMeta, errormdl.Wrap("fail to get media: " + err.Error())
		}
		rfile = fileFpCache.Set(f, nil)

		fpInfileIndex = nil
	} else {
		fpInfileIndex = rfile.InfileIndex
		f = rfile.file
	}

	rfile.lock.Lock()
	defer rfile.lock.Unlock()
	var metaData *gjson.Result
	dataByte, metaData, err = filepack.GetMediaFromPackFileUsingFp(f, fpInfileIndex, recordID)
	if err != nil {
		loggermdl.LogError(err)
		return dataByte, fileMeta, err
	}
	updatedInfileIndex := metaData.Get("infileIndex").String()
	ui := gjson.Parse(updatedInfileIndex)
	fileFpCache.Set(f, &ui)
	fileMeta = metaData.Get("requiredData")
	return dataByte, gjson.Parse(fileMeta.String()), nil
}

// func saveDataInMixedBucket(fdb *FDB, filePath string, index *Index, rs *gjson.Result) error {
// 	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 	bucket := fdb.buckets[bucketID]
// 	fileType := rs.Get("fileType").String()
// 	if len(fileType) == 0 {
// 		return errormdl.Wrap("please specify fileType")
// 	}
// 	isFilePresent := filemdl.FileAvailabilityCheck(filePath)

// 	infileIndex, ok := bucket.InFileIndexMap[fileType]
// 	if !ok {
// 		loggermdl.LogError("infileIndex for specified fileType not found: ", fileType)
// 		return errormdl.Wrap("infileIndex for specified fileType not found: " + fileType)
// 	}

// 	err := filepack.AddFileInPackFile(filePath, infileIndex.IndexFields, rs)
// 	if err != nil {
// 		loggermdl.LogError("fail to add fdb index entry in file: ", err)
// 		return errormdl.Wrap("fail to write data in file")
// 	}
// 	// for adding fdb index data
// 	if !isFilePresent {
// 		err = addFDBIndexEntryFile(filePath, bucket, index.IndexFields, rs)
// 	}
// 	return err
// }

func saveDataInAppendBucketUsingFp(f *os.File, bucket *Bucket, rs *gjson.Result, secParams securitymdl.FDBSecParams) error {

	// loggermdl.LogError("isLazyWriterEnabled", isLazyWriterEnabled)
	if bucket.EnableLazy {
		return saveDatInLazyBucket(bucket, f, rs, secParams)
	}

	//  else

	var (
		dataBytes   = []byte(rs.String())
		_, fileName = filepath.Split(f.Name())
		err         error
	)

	if secParams.EnableSecurity {
		key, err := securitymdl.GenSecKeyBytes(fileName, rs)
		if err != nil {
			loggermdl.LogError("failed to generate security key for append bucket: ", err)
			return err
		}

		dataBytes, err = encryptWithCompression(dataBytes, secParams.EnableCompression, key)
		if err != nil {
			loggermdl.LogError("failed to encrypt data for append bucket: ", err)
			return err
		}
	}

	dataBytes = []byte(string(dataBytes) + lineBreak)
	_, _, err = filemdl.AppendDataToFile(f.Name(), dataBytes, true)
	return err
}

// func addFDBIndexEntryFile(filePath string, bucket *Bucket, indexFields []IndexField, rs *gjson.Result) error {
// 	infileIndexFields := []filepack.InFileIndexField{}
// 	for _, field := range indexFields {
// 		infileIndexField := filepack.InFileIndexField{}
// 		infileIndexField.FieldName = field.FieldName
// 		infileIndexField.Query = field.Query
// 		infileIndexFields = append(infileIndexFields, infileIndexField)
// 	}
// 	// adding FDBIndex as infileIndex
// 	FDBIndexInFileIndexMap := filepack.InFileIndex{
// 		FileType:    FileTypeFDBIndex,
// 		IndexFields: infileIndexFields,
// 	}
// 	bucket.SetInFileIndex(FDBIndexInFileIndexMap)
// 	// create infile index data
// 	newData, err := filepack.CreateIndexJSON(infileIndexFields, rs)
// 	if err != nil {
// 		loggermdl.LogError(err)
// 		return err
// 	}
// 	newData, _ = sjson.Set(newData, "fileType", FileTypeFDBIndex)
// 	newDataObj := gjson.Parse(newData)
// 	return filepack.AddFileInPackFile(filePath, infileIndexFields, &newDataObj)
// }

func addFDBIndexEntryFileUsingFp(f *os.File, bucket *Bucket, indexFields []IndexField, rs *gjson.Result, secParams securitymdl.FDBSecParams) error {
	infileIndexFields := []filepack.InFileIndexField{}
	for _, field := range indexFields {
		infileIndexField := filepack.InFileIndexField{}
		infileIndexField.FieldName = field.FieldName
		infileIndexField.Query = field.Query
		infileIndexFields = append(infileIndexFields, infileIndexField)
	}
	// adding FDBIndex as infileIndex
	FDBIndexInFileIndexMap := filepack.InFileIndex{
		FileType:    FileTypeFDBIndex,
		IndexFields: infileIndexFields,
	}
	bucket.SetInFileIndex(FDBIndexInFileIndexMap)
	// create infile index data
	newData, err := filepack.CreateIndexJSON(infileIndexFields, rs)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	newData, _ = sjson.Set(newData, "fileType", FileTypeFDBIndex)

	fieldQuery := securitymdl.GetFDBSecOptions().FieldQuery
	if fieldQuery != "" {
		newData, _ = sjson.Set(newData, fieldQuery, rs.Get(fieldQuery).Value()) // set data required user defined query to RS. This is required for encryption.
	}
	newDataObj := gjson.Parse(newData)
	return filepack.AddFileInPackFileUsingFp(f, infileIndexFields, &newDataObj, secParams)
}

// func updateDataInNormalBucket(fdb *FDB, filePaths []string, rs *gjson.Result) (*gjson.Result, []string, []error) {
// 	resultStr := "[]"
// 	errList := []error{}
// 	filesUpdated, updatedCnt := make([]string, len(filePaths)), 0
// 	for _, filePath := range filePaths {
// 		path := filepath.Join(fdb.DBPath, filePath)
// 		data, err := updateNormalFileData(path, *rs)
// 		if err != nil {
// 			// loggermdl.LogError("failed to update data in file: " + path + " : " + err.Error())
// 			errList = append(errList, errormdl.Wrap("failed to update data in file: "+path+" : "+err.Error()))
// 			continue
// 		}
// 		if filePath != "" {
// 			// filesUpdated = append(filesUpdated, filePath)
// 			filesUpdated[updatedCnt] = filePath
// 			updatedCnt++
// 			resultStr, _ = sjson.Set(resultStr, "-1", data.Value())
// 		} else {
// 			loggermdl.LogError("invalid path for update - ", filePath)
// 		}
// 	}
// 	result := gjson.Parse(resultStr)
// 	return &result, filesUpdated[:updatedCnt], errList
// }

func updateDataInNormalBucketUsingFp(fdb *FDB, bucket *Bucket, fps []*os.File, rs *gjson.Result) (*gjson.Result, []*os.File, []error) {
	resultStr := "[]"
	errList := []error{}
	updatedfps := make([]*os.File, 0)
	// updatedCnt := 0
	var data *gjson.Result
	var err error
	lazyEnable := bucket.EnableLazy
	for _, fp := range fps {
		if lazyEnable {
			data, err = updateDataInLazyBucket(bucket, fp, rs, securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression})
			if err != nil {
				// loggermdl.LogError("failed to update data in file: " + path + " : " + err.Error())
				errList = append(errList, errormdl.Wrap("failed to update data in file: "+fp.Name()+" : "+err.Error()))
				continue
			}
		} else {
			data, err = updateNormalFileDataUsingFp(fp, bucket, rs, securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression})
			if err != nil {
				// loggermdl.LogError("failed to update data in file: " + path + " : " + err.Error())
				errList = append(errList, errormdl.Wrap("failed to update data in file: "+fp.Name()+" : "+err.Error()))
				continue
			}
		}
		// filePath := filepath.Join(strings.TrimPrefix(fp.Name(), fdb.DBPath))
		// path := filepath.Join(fdb.DBPath, filePath)
		// if filePath != "" {
		updatedfps = append(updatedfps, fp)
		// filesUpdated = append(filesUpdated, filePath)
		// filesUpdated[updatedCnt] = filePath
		// updatedCnt++
		resultStr, _ = sjson.Set(resultStr, "-1", data.Value())
		// } else {
		// 	loggermdl.LogError("invalid path for update - ", filePath)
		// }
	}
	result := gjson.Parse(resultStr)
	return &result, updatedfps, errList
}

// func updateNormalFileData(filePath string, rs gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, error) {
// 	data, err := readNormalFile(filePath)
// 	if err != nil {
// 		return nil, err
// 	}
// 	existingDataStr := string(data)
// 	rs.ForEach(func(key, val gjson.Result) bool {
// 		// updating existing data
// 		existingDataStr, _ = sjson.Set(existingDataStr, key.String(), val.Value())
// 		return true
// 	})

// 	updatedData := gjson.Parse(existingDataStr)
// 	err = saveDataInNormalBucket(filePath, &updatedData, secParams)
// 	// err = filemdl.FastWriteFile(filePath, []byte(existingDataStr), true, false, true)
// 	if err != nil {
// 		return nil, err
// 	}
// 	return &updatedData, nil
// }

func updateNormalFileDataUsingFp(fp *os.File, bucket *Bucket, rs *gjson.Result, secParams securitymdl.FDBSecParams) (*gjson.Result, error) {
	data, err := readNormalFileUsingFp(fp, rs, secParams)
	if err != nil {
		return nil, err
	}
	existingDataStr := string(data)
	rs.ForEach(func(key, val gjson.Result) bool {
		// updating existing data
		existingDataStr, _ = sjson.Set(existingDataStr, key.String(), val.Value())
		return true
	})

	updatedData := gjson.Parse(existingDataStr)
	// fileInfo, err := fp.Stat()
	// if err != nil {
	// 	return nil, err
	// }
	err = fp.Truncate(0)
	if err != nil {
		return nil, err
	}
	_, err = fp.Seek(0, 0)
	if err != nil {
		return nil, err
	}
	err = saveDataInNormalBucketUsingFp(fp, bucket, &updatedData, secParams)
	// err = filemdl.FastWriteFile(filePath, []byte(existingDataStr), true, false, true)
	if err != nil {
		return nil, err
	}
	return &updatedData, nil
}

// func updateDataInFileIndexBucket(fdb *FDB, bucket *Bucket, filePaths []string, rs *gjson.Result, infileIndexQuery []string) (*gjson.Result, []string, []error) {
// 	// if bucket.BucketType == BucketTypePack {
// 	requestedFileType := rs.Get("fileType").String()
// 	if len(requestedFileType) == 0 {
// 		loggermdl.LogError("please specify fileType")
// 		return nil, []string{}, []error{errormdl.Wrap("please specify fileType")}
// 	}
// 	infileIndexQuery = append(infileIndexQuery, `#[fileType=="`+requestedFileType+`"]`)
// 	// }
// 	finalResultArray := []gjson.Result{}
// 	errList := []error{}
// 	dataUpdatedAtFilePaths := []string{}
// 	dataUpdatedAtFilePathsCnt := 0
// 	for _, path := range filePaths {
// 		filePath := filepath.Join(fdb.DBPath, path)
// 		resultArray, err := filepack.UpdateFilesInPackFile(filePath, infileIndexQuery, rs)
// 		if err != nil {
// 			errList = append(errList, err)
// 			continue
// 		}

// 		dataUpdatedAtFilePaths = append(dataUpdatedAtFilePaths, path)
// 		// dataUpdatedAtFilePaths[dataUpdatedAtFilePathsCnt] = path
// 		dataUpdatedAtFilePathsCnt += len(resultArray.Array())
// 		finalResultArray = append(finalResultArray, resultArray.Array()...)
// 	}

// 	resultListStr := "[]"
// 	for _, resultObj := range finalResultArray {
// 		resultListStr, _ = sjson.Set(resultListStr, "-1", resultObj.Value())
// 	}
// 	result := gjson.Parse(resultListStr)
// 	return &result, dataUpdatedAtFilePaths, errList
// }

func updateDataInFileIndexBucketUsingFp(fdb *FDB, bucket *Bucket, fpInfileIndexMap map[*os.File]*gjson.Result, rs *gjson.Result, infileIndexQuery []string) (*gjson.Result, map[*os.File]*gjson.Result, []error) {
	// if bucket.BucketType == BucketTypePack {
	requestedFileType := rs.Get("fileType").String()
	if len(requestedFileType) == 0 {
		loggermdl.LogError("please specify fileType")
		return nil, fpInfileIndexMap, []error{errormdl.Wrap("please specify fileType")}
	}
	infileIndexQuery = append(infileIndexQuery, `#[fileType=="`+requestedFileType+`"]`)
	// }
	finalResultArray := []gjson.Result{}
	errList := []error{}
	updatedfpInfileIndexMap := make(map[*os.File]*gjson.Result)
	// dataUpdatedAtFilePaths := []string{}

	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}
	for fp, infileIndex := range fpInfileIndexMap {
		// filePath := filepath.Join(fdb.DBPath, path)
		resultArray, updatedInfileIndex, err := filepack.UpdateFileInPackFileUsingFp(fp, infileIndexQuery, infileIndex, rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			errList = append(errList, err)
			continue
		}
		updatedfpInfileIndexMap[fp] = updatedInfileIndex
		// dataUpdatedAtFilePaths = append(dataUpdatedAtFilePaths, path)
		// dataUpdatedAtFilePaths[dataUpdatedAtFilePathsCnt] = path
		// dataUpdatedAtFilePathsCnt += len(resultArray.Array())
		finalResultArray = append(finalResultArray, resultArray.Array()...)
	}

	resultListStr := "[]"
	for _, resultObj := range finalResultArray {
		resultListStr, _ = sjson.Set(resultListStr, "-1", resultObj.Value())
	}
	result := gjson.Parse(resultListStr)
	return &result, updatedfpInfileIndexMap, errList
}

// func readDataFromNormalBucket(fdb *FDB, paths []string) (*gjson.Result, error) {
// 	data, err := fdb.readNormalFiles(paths)
// 	if err != nil {
// 		loggermdl.LogError(err)
// 		return nil, err
// 	}
// 	return data, nil
// }

// func readDataFromNormalBucketUsingFp(fps []*os.File, bucket Bucket) (*gjson.Result, error) {
// 	data, err := readNormalFilesUsingFp(fps)
// 	if err != nil {
// 		loggermdl.LogError(err)
// 		return nil, err
// 	}
// 	return data, nil
// }

// ReindexOnSpecialBucket - ReindexOnSpecialBucket
func ReindexOnSpecialBucket(fdbName string, indexID string, rs gjson.Result) error {
	fdb, err := GetFDBInstance(fdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", fdbName)
		return errormdl.Wrap(("fdb instance not found: " + fdbName))
	}

	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: ", indexID)
		return errormdl.Wrap("index not found: " + indexID)
	}
	// find path to start file walk
	pathToStartWalk := fdb.DBPath
	for _, bucketID := range index.BucketSequence {
		bucket := fdb.buckets[bucketID]
		if bucket.IsDynamicName {
			break
		}
		pathToStartWalk = filepath.Join(pathToStartWalk, bucket.BucketNameQuery)
	}
	// bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	// bucket := fdb.buckets[bucketID]
	indexDataMap := make(map[string]string)
	if !filemdl.FileAvailabilityCheck(pathToStartWalk) {
		return errormdl.Wrap("invalid path: " + pathToStartWalk)
	}

	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}

	ferr := filemdl.Walk(pathToStartWalk, func(filePath string, info os.FileInfo, err error) error {
		if err != nil {
			loggermdl.LogError(err)
			return nil
		}

		if !info.IsDir() {
			result, err := filepack.GetDataFromPackFile(filePath, []string{`#[fileType==` + FileTypeFDBIndex + `]`}, &rs, secParams)
			if err != nil {
				loggermdl.LogError(err)
				return nil
			}
			data := gjson.Parse(result)
			data = data.Get("0")
			pathToSave := strings.TrimPrefix(filePath, fdb.DBPath+string(filepath.Separator))
			valString, _ := sjson.Delete(data.String(), "fileType")
			// val= gjson.Parse(valString)
			indexDataMap[pathToSave] = strings.TrimSpace(valString)
		}

		return nil
	})
	if ferr != nil {
		return ferr
	}
	return reCreateFDBIndexFile(fdb, index, indexDataMap)
}

func reCreateFDBIndexFile(fdb *FDB, index *Index, indexDataMap map[string]string) error {
	// create or replace index
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	index.indexObj.Close()
	indexFilePath := filepath.Join(fdb.DBPath, INDEXFOLDER, index.IndexID)
	if filemdl.FileAvailabilityCheck(indexFilePath) {
		err := filemdl.DeleteFile(indexFilePath)
		if err != nil {
			return err
		}
	}
	db, err := fdb.GetIndexDB(index)
	if err != nil {
		return err
	}
	index.indexObj = db
	err = index.indexObj.ReplaceIndex(index.IndexID, "*", fns...)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// update index file by reading all data and updating index file
	return index.indexObj.Update(func(tx *buntdb.Tx) error {

		for indexKey, indexData := range indexDataMap {
			_, _, err := tx.Set(indexKey, indexData, nil)
			if err != nil {
				loggermdl.LogError(err)
				return err
			}
		}

		return nil
	})
}

// LogFDBIndexFile -LogFDBIndexFile
func LogFDBIndexFile(indexFilePath string, index *Index) error {
	// dbPath := filepath.Join(fdbPath, INDEXFOLDER)
	// loggermdl.LogDebug("in log fdb index")
	dataToStore := ``

	err := index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			dataToStore = dataToStore + key + IndexKeyValSeperator + value + lineBreak
			// rsJSON := gjson.Parse("[" + value + "]")
			return true
		})
	})

	if err != nil {
		return err
	}
	_, fileName := filepath.Split(indexFilePath)
	dataByteToWriteRes, hashError := encryptData([]byte(dataToStore), fileName)
	if errormdl.CheckErr1(hashError) != nil {
		return errormdl.CheckErr1(hashError)
	}
	// dataByteToWriteRes := []byte(dataToStore)
	return filemdl.WriteFile(indexFilePath, dataByteToWriteRes, true, false)
}

// AddIndexEntriesInFile -AddIndexEntriesInFile
func AddIndexEntriesInFile(indexFilePath string, entries map[string]string) error {
	// dbPath := filepath.Join(fdbPath, INDEXFOLDER)
	// loggermdl.LogDebug("in log fdb index")
	dataToStore := ""
	for key, val := range entries {
		dataToStore = dataToStore + key + IndexKeyValSeperator + val + lineBreak
	}
	dataByteToWriteRes := []byte(dataToStore)
	var hashError error
	if len(dataByteToWriteRes) > 0 {
		_, fileName := filepath.Split(indexFilePath)
		dataByteToWriteRes, hashError = encryptData(dataByteToWriteRes, fileName)
		if errormdl.CheckErr1(hashError) != nil {
			return errormdl.CheckErr1(hashError)
		}
	}
	// dataByteToWriteRes := []byte(dataToStore)
	//creating
	return filemdl.WriteFile(indexFilePath, dataByteToWriteRes, true, false)
}

// GetKeyWithFileNameAndDefaultKey generates key using file name + Default key
func GetKeyWithFileNameAndDefaultKey(filePath string) ([]byte, error) {
	fileName := filepath.Base(filePath)
	fileNameBytes := []byte(fileName)
	fileNameBytes = append(fileNameBytes, defaultSecurityKey...)
	keyBytes, getHashError := hashmdl.Get128BitHash(fileNameBytes)
	if errormdl.CheckErr(getHashError) != nil {
		return nil, errormdl.CheckErr(getHashError)
	}
	return keyBytes[:], nil
}

// LoadFDBIndexFromFile -
func LoadFDBIndexFromFile(indexFilePath string, fdb *FDB, indexID string) error {
	index, found := fdb.GetFDBIndex(indexID)
	if !found {
		return errormdl.Wrap("index not found")
	}
	if !filemdl.FileAvailabilityCheck(indexFilePath) {
		return nil
	}
	fileData, err := filemdl.FastReadFile(indexFilePath)
	if err != nil {
		loggermdl.LogError("failed to load FDB index from: ", indexFilePath)
		return err
	}

	_, fileName := filepath.Split(indexFilePath)
	fileData, err = decryptData(fileData, fileName)
	if err != nil {
		loggermdl.LogError("failed to decrypt FDB index data: ", err)
		return errormdl.Wrap("failed to decrypt FDB index data: " + err.Error())
	}
	data := string(fileData)
	indexRecords := strings.Split(data, lineBreak)
	indexDataMap := make(map[string]string)
	for _, indexRecord := range indexRecords {
		indexValues := strings.Split(indexRecord, IndexKeyValSeperator)
		if len(indexValues) == 2 {
			indexDataMap[indexValues[0]] = indexValues[1]
		}
	}
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	// update index file by reading all data and updating index file
	return index.indexObj.Update(func(tx *buntdb.Tx) error {
		for key, val := range indexDataMap {
			_, _, err := tx.Set(key, val, nil)
			if err != nil {
				return err
			}
		}
		return nil
	})
}

// func deleteDataFromNormalBucket(fdbPath string, index *Index, paths []string) (recordsDeleted int, dataDeletedFromPaths []string, errList []error) {
// 	for _, path := range paths {
// 		filePath := filepath.Join(fdbPath, path)
// 		// delete file
// 		err := filemdl.DeleteFile(filePath)
// 		// err = deleteNormalFile(filePath)
// 		if err != nil {
// 			errList = append(errList, errormdl.Wrap("unable to delete file : "+err.Error()))
// 			continue
// 		}
// 		dataDeletedFromPaths = append(dataDeletedFromPaths, path)
// 		recordsDeleted++
// 	}

// 	return
// }

func deleteDataFromNormalBucketUsingFp(index *Index, bucket *Bucket, fps []*os.File) (recordsDeleted int, dataDeletedFromPaths []string, errList []error) {
	lazyEnable := bucket.EnableLazy
	for _, fp := range fps {
		if lazyEnable {
			err := deleteDataFromLazyBucket(bucket, fp)
			if err != nil {
				errList = append(errList, errormdl.Wrap("unable to delete file from lazy cache : "+err.Error()))
			}
		}
		// filePath := filepath.Join(fdbPath, path)
		// delete file
		if filemdl.FileAvailabilityCheck(fp.Name()) {
			// path := fp.Name()

			err := fp.Truncate(0)
			if err != nil {
				errList = append(errList, errormdl.Wrap("unable to delete file : "+err.Error()))
				continue
			}

			if _, err := fp.Seek(0, io.SeekStart); err != nil {
				errList = append(errList, errormdl.Wrap("unable to delete file : "+err.Error()))
				continue
			}
			err = fp.Close()
			if err != nil {
				errList = append(errList, errormdl.Wrap("unable to delete file : "+err.Error()))
				continue
			}
			// err = filemdl.DeleteFile(fp.Name())
			// // err = deleteNormalFile(filePath)
			// if err != nil {
			// 	errList = append(errList, errormdl.Wrap("unable to delete file : "+err.Error()))
			// 	continue
			// }
			dataDeletedFromPaths = append(dataDeletedFromPaths, fp.Name())
			recordsDeleted++

			// delete file from cache
			fileFpCache.Delete(fp.Name())
		}
	}

	return
}

// func deleteNormalFile(filePath string) error {
// 	return filemdl.DeleteFile(filePath)
// }

// func deleteDataFromInFileIndexBucket(fdbPath string, bucket *Bucket, paths []string, rs *gjson.Result, infileIndexQueries []string) (recordsDeleted int, dataDeletedFromPath []string, errList []error) {

// 	fileType := rs.Get("fileType").String()
// 	if len(fileType) == 0 {
// 		loggermdl.LogError("fileType value not provided")
// 		return recordsDeleted, dataDeletedFromPath, []error{errormdl.Wrap("please specify fileType")}
// 	}

// 	_, ok := bucket.InFileIndexMap[fileType]
// 	if !ok {
// 		loggermdl.LogError("infileIndex for specified fileType not found")
// 		return recordsDeleted, dataDeletedFromPath, []error{errormdl.Wrap("infileIndex for specified fileType not found")}
// 	}

// 	infileIndexQueries = append(infileIndexQueries, `#[fileType=="`+fileType+`"]`)
// 	noDataFoundCnt := 0

// 	for _, filePath := range paths {
// 		path := filepath.Join(fdbPath, filePath)
// 		recordsDeletedCnt, err := filepack.DeletDataFromPackFile(path, infileIndexQueries)
// 		if err == ErrNoDataFound {
// 			noDataFoundCnt++
// 			continue
// 		}
// 		if err != nil {
// 			errList = append(errList, err)
// 			continue
// 		}
// 		dataDeletedFromPath = append(dataDeletedFromPath, filePath)
// 		recordsDeleted += recordsDeletedCnt
// 	}

// 	if noDataFoundCnt == len(paths) {
// 		errList = []error{ErrNoDataFound}
// 	}
// 	return
// }
func deleteDataFromPackBucketUsingFp(bucket *Bucket, fpInfileIndexMap map[*os.File]*gjson.Result, rs *gjson.Result, infileIndexQueries []string) (recordsDeleted int, updatedFpInfileIndexMap map[*os.File]*gjson.Result, errList []error) {
	// updatedFpInfileIndexMap := make(map[*os.File]*gjson.Result)
	updatedFpInfileIndexMap = make(map[*os.File]*gjson.Result)
	fileType := rs.Get("fileType").String()
	if len(fileType) == 0 {
		loggermdl.LogError("fileType value not provided")
		return recordsDeleted, updatedFpInfileIndexMap, []error{errormdl.Wrap("please specify fileType")}
	}

	_, ok := bucket.InFileIndexMap[fileType]
	if !ok {
		loggermdl.LogError("infileIndex for specified fileType not found")
		return recordsDeleted, updatedFpInfileIndexMap, []error{errormdl.Wrap("infileIndex for specified fileType not found")}
	}

	infileIndexQueries = append(infileIndexQueries, `#[fileType=="`+fileType+`"]`)
	noDataFoundCnt := 0
	for fp, infileIndex := range fpInfileIndexMap {
		// path := filepath.Join(fdbPath, filePath)
		recordsDeletedCnt, updatedInfileIndex, err := filepack.DeletDataFromPackFileUsingFp(fp, infileIndex, infileIndexQueries)
		if err != nil && (err.Error() == ErrNoDataFound.Error()) {
			noDataFoundCnt++
			continue
		}
		if err != nil {
			errList = append(errList, err)
			continue
		}
		// dataDeletedFromPath = append(dataDeletedFromPath, filePath)
		recordsDeleted += recordsDeletedCnt
		updatedFpInfileIndexMap[fp] = updatedInfileIndex
	}

	if noDataFoundCnt == len(fpInfileIndexMap) {
		errList = []error{ErrNoDataFound}
	}
	return
}

// // SaveDataInFDBOld -SaveDataInFDBOld
// func SaveDataInFDBOld(dbName, indexID string, rs *gjson.Result) error {
// 	fdb, err := GetFDBInstance(dbName)
// 	if err != nil {
// 		return err
// 	}
// 	index, ok := fdb.GetFDBIndex(indexID)
// 	if !ok {
// 		loggermdl.LogError("INDEX not found: " + indexID)
// 		return errormdl.Wrap("INDEX not found: " + indexID)
// 	}
// 	path, err := fdb.resolveIndex(index, rs)
// 	if errormdl.CheckErr(err) != nil {
// 		loggermdl.LogError(err)
// 		return errormdl.CheckErr(err)
// 	}

// 	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 	bucket, ok := fdb.buckets[bucketID]
// 	if !ok {
// 		loggermdl.LogError("Bucket not found: " + bucketID)
// 		return errormdl.Wrap("Bucket not found: " + bucketID)
// 	}
// 	filePath := filepath.Join(fdb.DBPath, path)
// 	// check is infile index present
// 	// if yes
// 	switch bucket.BucketType {
// 	case BucketTypeSimple:
// 		err := saveDataInNormalBucket(filePath, rs)
// 		if errormdl.CheckErr(err) != nil {
// 			loggermdl.LogError(err)
// 			return err
// 		}
// 	case BucketTypePack, BucketTypeMixed:
// 		err := saveDataInPackBucket(fdb, filePath, index, rs)
// 		if errormdl.CheckErr(err) != nil {
// 			loggermdl.LogError(err)
// 			return errormdl.CheckErr(err)
// 		}
// 	case BucketTypeAppend:
// 		err := saveDataInAppendBucket(filePath, rs, bucket)
// 		if errormdl.CheckErr(err) != nil {
// 			loggermdl.LogError(err)
// 			return errormdl.CheckErr(err)
// 		}
// 	default:
// 		return errormdl.Wrap("Please provide valid bucket type")
// 	}
// 	// loggermdl.LogDebug("index", *index)
// 	json, err := createIndexJSON(index, rs)
// 	if err != nil {
// 		loggermdl.LogError(err)
// 		return err
// 	}

// 	// path = path + ".json"
// 	err = index.indexObj.Update(func(tx *buntdb.Tx) error {
// 		// optimization : if data not changed then dont update
// 		_, _, err := tx.Set(path, json, nil)
// 		if err != nil {
// 			loggermdl.LogError(err)
// 			return err
// 		}
// 		return nil
// 	})
// 	// TODO: Currently index data is overwritten by new data.
// 	if isLazyWriterEnabled {
// 		err = UpdateIndexLazyObjectInCache(indexID, index)
// 		if err != nil {
// 			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
// 			return errormdl.Wrap("failed to update index data in lazy writer cache")
// 		}
// 	}
// 	return nil
// }

// // ReadDataFromFDBOld - return records with matching query
// func ReadDataFromFDBOld(dbName, indexID string, rs *gjson.Result, query []string, infileIndexQuery []string) (*gjson.Result, error) {
// 	fdb, err := GetFDBInstance(dbName)
// 	if err != nil {
// 		return nil, err
// 	}
// 	index, ok := fdb.GetFDBIndex(indexID)
// 	if !ok {
// 		loggermdl.LogError("INDEX not found: " + indexID)
// 		return nil, errormdl.Wrap("INDEX not found: " + indexID)
// 	}
// 	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 	bucket, ok := fdb.buckets[bucketID]
// 	if !ok {
// 		loggermdl.LogError("Bucket not found: " + bucketID)
// 		return nil, errormdl.Wrap("Bucket not found: " + bucketID)
// 	}
// 	filePaths := []string{}
// 	// loggermdl.LogDebug(query)
// 	index.indexObj.View(func(tx *buntdb.Tx) error {
// 		return tx.Ascend(index.IndexID, func(key, value string) bool {
// 			rsJSON := gjson.Parse("[" + value + "]")
// 			for i := 0; i < len(query); i++ {
// 				rsJSON = rsJSON.Get(query[i] + "#")
// 			}
// 			if rsJSON.Value() != nil {
// 				filePaths = append(filePaths, key)
// 			}
// 			return true
// 		})
// 	})
// 	resultArray := "[]"
// 	resultToReturn := gjson.Parse(resultArray)
// 	if len(filePaths) == 0 {
// 		loggermdl.LogError("files not found")
// 		return &resultToReturn, nil
// 	}
// 	if bucket.BucketType == BucketTypePack || bucket.BucketType == BucketTypeMixed {
// 		for _, path := range filePaths {
// 			filePath := filepath.Join(fdb.DBPath, path)
// 			requestedFileType := rs.Get("fileType").String()
// 			if len(requestedFileType) == 0 {
// 				return &resultToReturn, errormdl.Wrap("please specify fileType")
// 			}
// 			// bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 			// bucket := fdb.buckets[bucketID]
// 			infileIndexQuery = append(infileIndexQuery, `#[fileType==`+requestedFileType+`]`)
// 			result, err := filepack.GetDataFromPackFile(filePath, infileIndexQuery)
// 			if err != nil {
// 				loggermdl.LogError(err)
// 				return &resultToReturn, err
// 			}
// 			// loggermdl.LogDebug("result", result)
// 			for _, val := range gjson.Parse(result).Array() {
// 				resultArray, _ = sjson.Set(resultArray, "-1", val.Value())
// 			}
// 		}
// 		resultToReturn = gjson.Parse(resultArray)
// 		return &resultToReturn, nil
// 	} else if bucket.BucketType == BucketTypeSimple {
// 		return readDataFromNormalBucket(fdb, filePaths)
// 	} else {
// 		return nil, errormdl.Wrap("Operation not available on bucket type " + bucket.BucketType)
// 	}
// 	// loggermdl.LogError("resultToReturn - ", resultToReturn)

// }

// // UpdateDataInFDBOld - UpdateDataInFDBOld
// func UpdateDataInFDBOld(dbName, indexID string, rs *gjson.Result, query []string, infileIndexQuery []string) (*gjson.Result, []error) {
// 	var errList []error
// 	fdb, err := GetFDBInstance(dbName)
// 	if err != nil {
// 		return nil, errList
// 	}
// 	index, ok := fdb.GetFDBIndex(indexID)
// 	if !ok {
// 		loggermdl.LogError("INDEX not found: " + indexID)
// 		return nil, []error{errormdl.Wrap("INDEX not found: " + indexID)}
// 	}
// 	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 	bucket, ok := fdb.buckets[bucketID]
// 	if !ok {
// 		loggermdl.LogError("Bucket not found: " + bucketID)
// 		return nil, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
// 	}

// 	filePaths := []string{}
// 	indexKeyValMap := make(map[string]string)
// 	// loggermdl.LogDebug(query)
// 	index.indexObj.View(func(tx *buntdb.Tx) error {
// 		return tx.Ascend(index.IndexID, func(key, value string) bool {
// 			rsJSON := gjson.Parse("[" + value + "]")
// 			for i := 0; i < len(query); i++ {
// 				rsJSON = rsJSON.Get(query[i] + "#")
// 			}
// 			if rsJSON.Value() != nil {
// 				filePaths = append(filePaths, key)
// 				indexKeyValMap[key] = rsJSON.Array()[0].String()
// 			}
// 			return true
// 		})
// 	})

// 	// loggermdl.LogDebug("filePaths", filePaths)
// 	resultToReturn := gjson.Result{}

// 	if len(filePaths) == 0 {
// 		loggermdl.LogError("files not found")
// 		return nil, []error{ErrNoDataFound}
// 	}
// 	result := &gjson.Result{}
// 	var dataUpdatedAtFilePaths []string
// 	switch bucket.BucketType {
// 	case BucketTypeSimple:
// 		result, dataUpdatedAtFilePaths, errList = updateDataInNormalBucket(fdb, filePaths, rs)
// 	case BucketTypePack, BucketTypeMixed:
// 		result, dataUpdatedAtFilePaths, errList = updateDataInFileIndexBucket(fdb, bucket, filePaths, rs, infileIndexQuery)
// 	default:
// 		loggermdl.LogError("invalid bucket type")
// 		return nil, []error{errormdl.Wrap("invalid bucket type - " + bucket.BucketType)}
// 	}

// 	if len(dataUpdatedAtFilePaths) == 0 {
// 		loggermdl.LogError("no data found to update")
// 		return &resultToReturn, []error{ErrNoDataFound}
// 	}
// 	// path = path + ".json"
// 	index.indexObj.Update(func(tx *buntdb.Tx) error {
// 		// optimization : if data not changed then dont update
// 		for _, path := range dataUpdatedAtFilePaths {
// 			json := indexKeyValMap[path]
// 			updatedJSON, err := updateIndexJSON(index, json, rs)
// 			if err != nil {
// 				errList = append(errList, err)
// 			}
// 			_, _, err = tx.Set(path, updatedJSON, nil)
// 			if err != nil {
// 				errList = append(errList, err)
// 			}
// 		}
// 		return nil
// 	})

// 	// TODO: Currently index data is overwritten by new data.
// 	if isLazyWriterEnabled {
// 		err = UpdateIndexLazyObjectInCache(indexID, index)
// 		if err != nil {
// 			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
// 			return nil, []error{errormdl.Wrap("failed to update index data in lazy writer cache")}
// 		}
// 	}

// 	return result, errList
// }

// // DeleteDataFromFDBOld -
// func DeleteDataFromFDBOld(dbName string, indexID string, rs *gjson.Result, queries []string, infileIndexQueries []string) (recordsDeleted int, errList []error) {
// 	// errList := make([]error, 0)

// 	fdb, err := GetFDBInstance(dbName)
// 	if err != nil {
// 		loggermdl.LogError("fdb instance not found for: ", dbName)
// 		return recordsDeleted, []error{err}
// 	}
// 	index, ok := fdb.GetFDBIndex(indexID)
// 	if !ok {
// 		loggermdl.LogError("index not found: " + indexID)
// 		return recordsDeleted, []error{errormdl.Wrap("index not found: " + indexID)}
// 	}
// 	filePaths := []string{}
// 	// loggermdl.LogDebug(query)
// 	index.indexObj.View(func(tx *buntdb.Tx) error {
// 		return tx.Ascend(index.IndexID, func(key, value string) bool {
// 			// loggermdl.LogDebug(key, value)
// 			rsJSON := gjson.Parse("[" + value + "]")
// 			for i := 0; i < len(queries); i++ {
// 				rsJSON = rsJSON.Get(queries[i] + "#")
// 			}
// 			if rsJSON.Value() != nil {
// 				filePaths = append(filePaths, key)
// 			}
// 			return true
// 		})
// 	})
// 	if len(filePaths) == 0 {
// 		loggermdl.LogError("no data found to delete")
// 		return recordsDeleted, []error{ErrNoDataFound}
// 	}
// 	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
// 	bucket, ok := fdb.buckets[bucketID]
// 	if !ok {
// 		loggermdl.LogError("Bucket not found: " + bucketID)
// 		return recordsDeleted, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
// 	}
// 	dataDeletedFromPaths := []string{}
// 	if bucket.BucketType == BucketTypeSimple {
// 		recordsDeleted, dataDeletedFromPaths, errList = deleteDataFromNormalBucket(fdb.DBPath, index, filePaths)
// 	} else if bucket.BucketType == BucketTypePack || bucket.BucketType == BucketTypeMixed {
// 		recordsDeleted, dataDeletedFromPaths, errList = deleteDataFromInFileIndexBucket(fdb.DBPath, bucket, filePaths, rs, infileIndexQueries)
// 	} else {
// 		return recordsDeleted, []error{errormdl.Wrap("Operation not allowed on bucket type: " + bucket.BucketType)}
// 	}
// 	// loggermdl.LogError("dataDeletedFromPaths", dataDeletedFromPaths)
// 	if bucket.BucketType == BucketTypeSimple {
// 		index.indexObj.Update(func(tx *buntdb.Tx) error {
// 			// optimization : if data not changed then dont update
// 			for _, path := range dataDeletedFromPaths {
// 				_, err = tx.Delete(path)
// 				if err != nil {
// 					errList = append(errList, err)
// 				}
// 			}
// 			return nil
// 		})
// 	}

// 	if isLazyWriterEnabled {

// 		err := UpdateIndexLazyObjectInCache(index.IndexID, index)
// 		if err != nil {
// 			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
// 			return 0, []error{errormdl.Wrap("failed to update index data in lazy writer cache")}
// 		}
// 	}
// 	return recordsDeleted, errList
// }

// GetBucketByIndexID - return bucket by specified indexID
func GetBucketByIndexID(dbName, indexID string) (Bucket, error) {
	fdb, err := GetFDBInstance(dbName)
	bucket := Bucket{}
	if err != nil {
		loggermdl.LogError("fdb instance not found for: ", dbName)
		return bucket, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: " + indexID)
		return bucket, errormdl.Wrap("index not found: " + indexID)
	}

	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucketPtr, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return bucket, errormdl.Wrap("Bucket not found: " + bucketID)
	}
	return *bucketPtr, nil
}

func compressData(data []byte) ([]byte, error) {
	var hashError error
	dataByteToWriteRes, hashError := filemdl.ZipBytes(data)
	if errormdl.CheckErr2(hashError) != nil {
		return data, errormdl.CheckErr2(hashError)
	}
	return dataByteToWriteRes, nil
}
func decompressData(data []byte) ([]byte, error) {
	var hashError error
	dataByteToWriteRes, hashError := filemdl.UnZipBytes(data)
	if errormdl.CheckErr2(hashError) != nil {
		return data, errormdl.CheckErr2(hashError)
	}
	return dataByteToWriteRes, nil
}

func encryptData(data []byte, fileName string) (dataOut []byte, err error) {
	key, err := getSecurityKey(fileName)
	if err != nil {
		return dataOut, err
	}
	dataOut, err = compressData(data)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	dataOut, err = securitymdl.AESEncrypt(dataOut, key)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	return
}

func encryptWithCompression(data []byte, compress bool, encKey []byte) (res []byte, err error) {
	if compress {
		res, err = compressData(data)
		if err != nil {
			return
		}
	}

	return securitymdl.AESEncrypt(res, encKey)
}

func decryptwithDecompression(data []byte, deCompress bool, encKey []byte) (res []byte, err error) {
	res, err = securitymdl.AESDecrypt(data, encKey)
	if err != nil {
		return
	}

	if deCompress {
		return decompressData(res)
	}

	return
}

func decryptData(data []byte, fileName string) (dataOut []byte, err error) {

	key, err := getSecurityKey(fileName)
	if err != nil {
		return dataOut, err
	}

	dataOut, err = securitymdl.AESDecrypt(data, key)
	if errormdl.CheckErr1(err) != nil {
		return
	}

	dataOut, err = decompressData(dataOut)
	if errormdl.CheckErr1(err) != nil {
		return
	}
	return
}

func getSecurityKey(fileName string) ([]byte, error) {
	var key []byte
	var err error
	securityKeyGenFunc := securitymdl.GetSecurityKeyGeneratorFunc()
	if securityKeyGenFunc != nil {
		key, err = securityKeyGenFunc(fileName)
	} else {
		key, err = GetKeyWithFileNameAndDefaultKey(fileName)
	}
	return key, err
}

// SaveDataInFDB -
func SaveDataInFDB(dbName, indexID string, rs *gjson.Result) error {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return errormdl.Wrap("INDEX not found: " + indexID)
	}
	path, err := fdb.resolveIndex(index, rs)
	if errormdl.CheckErr(err) != nil {
		loggermdl.LogError(err)
		return errormdl.CheckErr(err)
	}
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return errormdl.Wrap("Bucket not found: " + bucketID)
	}
	prevVal, err := GetFDBIndexEntryByPath(index, path)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	filePath := filepath.Join(fdb.DBPath, path)
	var fp *os.File
	var inFileIndexData *gjson.Result
	rfile, err := fileFpCache.Get(filePath)
	if err != nil {
		fp, err = openFile(filePath)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		rfile = fileFpCache.Set(fp, nil)
	} else {
		fp = rfile.file
		inFileIndexData = rfile.InfileIndex
	}
	rfile.lock.Lock()
	defer rfile.lock.Unlock()
	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}
	switch bucket.BucketType {
	case BucketTypeSimple:
		// err = saveDataInNormalBucketUsingFp(fp, rs)
		err = saveDataInNormalBucketUsingFp(fp, bucket, rs, secParams)
	case BucketTypePack, BucketTypeMixed:
		inFileIndexData, err = saveDataInPackBucketUsingFp(fdb, fp, index, inFileIndexData, rs)
		fileFpCache.Set(fp, inFileIndexData)
	case BucketTypeAppend:
		err := saveDataInAppendBucketUsingFp(fp, bucket, rs, secParams)
		if errormdl.CheckErr(err) != nil {
			loggermdl.LogError(err)
			return errormdl.CheckErr(err)
		}
	default:
		return errormdl.Wrap("Please provide valid bucket type")
	}

	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// rfile, err = fileFpCache.Get(filePath)

	// json, err := createIndexJSON(index, rs)
	// if err != nil {
	// 	loggermdl.LogError(err)
	// 	return err
	// }

	// // path = path + ".json"
	// err = index.indexObj.Update(func(tx *buntdb.Tx) error {
	// 	// optimization : if data not changed then dont update
	// 	_, _, err := tx.Set(path, json, nil)
	// 	if err != nil {
	// 		loggermdl.LogError(err)
	// 		return err
	// 	}
	// 	return nil
	// })
	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return err

	}
	path = strings.TrimPrefix(fp.Name(), fdb.DBPath+string(filepath.Separator))
	prevVal, _ = sjson.Set(prevVal, "rowID", rowID)
	updatedJSON, err := updateIndexJSON(index, prevVal, rs)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	updatedJSONObj := gjson.Parse(updatedJSON)
	err = updateFDBIndexEntry(index, path, &updatedJSONObj)
	// _, _, err = tx.Set(path, updatedJSON, nil)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// _, err = addFDBIndexEntry(index, path, rs)
	// if err != nil {
	// 	loggermdl.LogError(err)
	// 	return err
	// }
	// TODO: Currently index data is overwritten by new data.
	if isLazyWriterEnabled {
		err = UpdateIndexLazyObjectInCache(indexID, index)
		if err != nil {
			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
			return errormdl.Wrap("failed to update index data in lazy writer cache")
		}
	}
	return nil
}

// UpdateDataInFDB -
func UpdateDataInFDB(dbName, indexID string, rs *gjson.Result, query []string, infileIndexQuery []string) (*gjson.Result, []error) {
	var errList []error
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return nil, []error{err}
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return nil, []error{errormdl.Wrap("INDEX not found: " + indexID)}
	}
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return nil, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
	}

	filePaths := []string{}
	indexKeyValMap := make(map[string]string)
	// loggermdl.LogDebug(query)
	err = index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend(index.IndexID, func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(query); i++ {
				rsJSON = rsJSON.Get(query[i] + "#")
			}
			if rsJSON.Value() != nil && rsJSON.IsArray() && len(rsJSON.Array()) > 0 {
				filePaths = append(filePaths, key)
				indexKeyValMap[key] = rsJSON.Array()[0].String()
			}
			return true
		})
	})

	if err != nil {
		return nil, []error{err}
	}
	// resultToReturn := gjson.Result{}

	if len(filePaths) == 0 {
		loggermdl.LogError("files not found")
		return nil, []error{ErrNoDataFound}
	}
	var fps []*os.File
	fpInfileIndexMap := make(map[*os.File]*gjson.Result)

	var rFiles []*File
	for _, filePath := range filePaths {
		filePath = filepath.Join(fdb.DBPath, filePath)
		rfile, err := fileFpCache.Get(filePath)
		if err != nil {
			fp, err := openFile(filePath)
			if err != nil {
				continue
			}
			// fileFpCache.Set(fp, nil)
			fps = append(fps, fp)
			fpInfileIndexMap[fp] = nil
			rfile = fileFpCache.Set(fp, nil)
		} else {
			fpInfileIndexMap[rfile.file] = rfile.InfileIndex
			fps = append(fps, rfile.file)
		}
		rfile.lock.Lock()
		rFiles = append(rFiles, rfile)
	}

	defer func() {
		for i := range rFiles {
			rFiles[i].lock.Unlock()
		}
	}()
	var result *gjson.Result
	// var dataUpdatedAtFilePaths []string
	var updatedFpInfileIndexMap map[*os.File]*gjson.Result
	var updatedFps []*os.File
	switch bucket.BucketType {
	case BucketTypeSimple:
		result, updatedFps, errList = updateDataInNormalBucketUsingFp(fdb, bucket, fps, rs)
		// optimization : if data not changed then dont update
		for _, fp := range updatedFps {
			path := strings.TrimPrefix(fp.Name(), fdb.DBPath+string(filepath.Separator))
			json := indexKeyValMap[path]
			updatedJSON, err := updateIndexJSON(index, json, rs)
			if err != nil {
				errList = append(errList, err)
			}
			updatedJSONObj := gjson.Parse(updatedJSON)
			_, err = addFDBIndexEntry(index, path, &updatedJSONObj)
			// _, _, err = tx.Set(path, updatedJSON, nil)
			if err != nil {
				errList = append(errList, err)
			}
		}

	case BucketTypePack, BucketTypeMixed:
		result, updatedFpInfileIndexMap, errList = updateDataInFileIndexBucketUsingFp(fdb, bucket, fpInfileIndexMap, rs, infileIndexQuery)
		for fp, infileIndex := range updatedFpInfileIndexMap {
			fileFpCache.Set(fp, infileIndex)
			path := strings.TrimPrefix(fp.Name(), fdb.DBPath+string(filepath.Separator))
			json := indexKeyValMap[path]
			updatedJSON, err := updateIndexJSON(index, json, rs)
			if err != nil {
				loggermdl.LogError(err)
				errList = append(errList, err)
			}
			updatedJSONObj := gjson.Parse(updatedJSON)
			err = updateFDBIndexEntry(index, path, &updatedJSONObj)
			// _, _, err = tx.Set(path, updatedJSON, nil)
			if err != nil {
				loggermdl.LogError(err)
				errList = append(errList, err)
			}
		}
	default:
		loggermdl.LogError("invalid bucket type")
		return nil, []error{errormdl.Wrap("invalid bucket type - " + bucket.BucketType)}
	}

	// if len(dataUpdatedAtFilePaths) == 0 {
	// 	loggermdl.LogError("no data found to update")
	// 	return &resultToReturn, []error{ErrNoDataFound}
	// }

	// path = path + ".json"

	// TODO: Currently index data is overwritten by new data.
	if isLazyWriterEnabled {
		err = UpdateIndexLazyObjectInCache(indexID, index)
		if err != nil {
			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
			return nil, []error{errormdl.Wrap("failed to update index data in lazy writer cache")}
		}
	}

	return result, errList
}

// DeleteDataFromFDB -
func DeleteDataFromFDB(dbName string, indexID string, rs *gjson.Result, queries []string, infileIndexQueries []string) (recordsDeletedCnt int, errList []error) {
	// errList := make([]error, 0)

	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found for: ", dbName)
		return recordsDeletedCnt, []error{err}
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("index not found: " + indexID)
		return recordsDeletedCnt, []error{errormdl.Wrap("index not found: " + indexID)}
	}
	filePaths, err := getFDBIndexEntry(index, queries)
	if err != nil {
		loggermdl.LogError(err)
		return recordsDeletedCnt, []error{err}
	}
	if len(filePaths) == 0 {
		loggermdl.LogError("no data found to delete")
		return recordsDeletedCnt, []error{ErrNoDataFound}
	}
	fps := make([]*os.File, 0)
	fpInfileIndexMap := make(map[*os.File]*gjson.Result)
	var rFiles []*File
	for _, filePath := range filePaths {
		filePath = filepath.Join(fdb.DBPath, filePath)
		rfile, err := fileFpCache.Get(filePath)
		if err != nil {
			fp, err := openFile(filePath)
			if err != nil {
				continue
			}
			fps = append(fps, fp)
			fpInfileIndexMap[fp] = nil
			rfile = fileFpCache.Set(fp, nil)
		} else {
			fpInfileIndexMap[rfile.file] = rfile.InfileIndex
			fps = append(fps, rfile.file)
		}

		rfile.lock.Lock()
		rFiles = append(rFiles, rfile)
	}

	defer func() {
		for i := range rFiles {
			rFiles[i].lock.Unlock()
		}
	}()
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return recordsDeletedCnt, []error{errormdl.Wrap("Bucket not found: " + bucketID)}
	}
	dataDeletedFromPaths := []string{}
	var updatedfpInfileIndexMap map[*os.File]*gjson.Result
	if bucket.BucketType == BucketTypeSimple {
		recordsDeletedCnt, dataDeletedFromPaths, errList = deleteDataFromNormalBucketUsingFp(index, bucket, fps)
		err = index.indexObj.Update(func(tx *buntdb.Tx) error {
			// optimization : if data not changed then dont update
			for _, path := range dataDeletedFromPaths {
				// loggermdl.LogError(path)
				path = strings.TrimPrefix(path, fdb.DBPath+string(filepath.Separator))
				_, err = tx.Delete(path)
				if err != nil {
					loggermdl.LogError(err)
					errList = append(errList, err)
				}
			}
			return nil
		})

		if err != nil {
			return recordsDeletedCnt, []error{err}
		}

	} else if bucket.BucketType == BucketTypePack || bucket.BucketType == BucketTypeMixed {
		recordsDeletedCnt, updatedfpInfileIndexMap, errList = deleteDataFromPackBucketUsingFp(bucket, fpInfileIndexMap, rs, infileIndexQueries)
		for fp, infileIndex := range updatedfpInfileIndexMap {
			fileFpCache.Set(fp, infileIndex)
		}
	} else {
		return recordsDeletedCnt, []error{errormdl.Wrap("Operation not allowed on bucket type: " + bucket.BucketType)}
	}
	// loggermdl.LogError("dataDeletedFromPaths", dataDeletedFromPaths)

	if isLazyWriterEnabled {

		err := UpdateIndexLazyObjectInCache(index.IndexID, index)
		if err != nil {
			loggermdl.LogError("failed to update index data in lazy writer cache - ", err)
			return 0, []error{errormdl.Wrap("failed to update index data in lazy writer cache")}
		}
	}
	return recordsDeletedCnt, errList
}

// ReadDataFromFDB - return records with matching query
func ReadDataFromFDB(dbName, indexID string, rs *gjson.Result, queries []string, infileIndexQuery []string) (*gjson.Result, error) {
	fdb, err := GetFDBInstance(dbName)
	if err != nil {
		return nil, err
	}
	index, ok := fdb.GetFDBIndex(indexID)
	if !ok {
		loggermdl.LogError("INDEX not found: " + indexID)
		return nil, errormdl.Wrap("INDEX not found: " + indexID)
	}
	bucketID := index.BucketSequence[len(index.BucketSequence)-1]
	bucket, ok := fdb.buckets[bucketID]
	if !ok {
		loggermdl.LogError("Bucket not found: " + bucketID)
		return nil, errormdl.Wrap("Bucket not found: " + bucketID)
	}
	// filePaths := []string{}
	// loggermdl.LogDebug(query)
	filePaths, err := getFDBIndexEntry(index, queries)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}
	resultToReturn := gjson.Parse("[]")
	if len(filePaths) == 0 {
		loggermdl.LogError("files not found")
		return &resultToReturn, nil
	}
	fps := make([]*os.File, 0)
	fpInfileIndexMap := make(map[*os.File]*gjson.Result)
	var rFiles []*File
	for _, filePath := range filePaths {
		filePath = filepath.Join(fdb.DBPath, filePath)
		rfile, err := fileFpCache.Get(filePath)
		if err != nil {
			fp, err := openFile(filePath)
			if err != nil {
				continue
			}
			// fileFpCache.Set(fp, nil)
			fps = append(fps, fp)
			fpInfileIndexMap[fp] = nil
			rfile = fileFpCache.Set(fp, nil)
		} else {
			fpInfileIndexMap[rfile.file] = rfile.InfileIndex
			fps = append(fps, rfile.file)
		}
		rfile.lock.Lock()
		rFiles = append(rFiles, rfile)
	}

	defer func() {
		for i := range rFiles {
			rFiles[i].lock.Unlock()
		}
	}()
	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}

	switch bucket.BucketType {
	case BucketTypeSimple:
		return readNormalFilesUsingFp(fps, bucket, rs, secParams)
	case BucketTypePack, BucketTypeMixed:
		requestedFileType := rs.Get("fileType").String()
		if len(requestedFileType) == 0 {
			return &resultToReturn, errormdl.Wrap("please specify fileType")
		}
		infileIndexQuery = append(infileIndexQuery, `#[fileType==`+requestedFileType+`]`)
		resultArray, updatedFpInfileIndexMap, err := readDataFromPackFiles(fpInfileIndexMap, infileIndexQuery, rs, secParams)
		if err != nil {
			loggermdl.LogError(err)
			return &resultToReturn, err
		}
		for fp, infileIndex := range updatedFpInfileIndexMap {
			fileFpCache.Set(fp, infileIndex)
		}
		resultToReturn = gjson.Parse(resultArray)
		return &resultToReturn, nil
	}
	return nil, errormdl.Wrap("Operation not available on bucket type: " + bucket.BucketType)
	// loggermdl.LogError("resultToReturn - ", resultToReturn)
}

func getFDBIndexEntry(index *Index, queries []string) ([]string, error) {
	filePaths := []string{}
	err := index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(queries); i++ {
				rsJSON = rsJSON.Get(queries[i] + "#")
			}
			if rsJSON.Value() != nil && rsJSON.IsArray() && len(rsJSON.Array()) > 0 {
				filePaths = append(filePaths, key)
			}
			return true
		})
	})
	return filePaths, err
}

func readDataFromPackFiles(filepathInfileIndexMap map[*os.File]*gjson.Result, infileIndexQuery []string, rs *gjson.Result, secParams securitymdl.FDBSecParams) (string, map[*os.File]*gjson.Result, error) {
	resultArray := "[]"
	// filepathInfileIndexMap := make(map[string]*gjson.Result)
	for f, infileIndexData := range filepathInfileIndexMap {
		result, updatedInfileIndexData, err := filepack.GetDataFromPackFileUsingFp(f, infileIndexData, infileIndexQuery, rs, secParams)
		if err != nil {
			return resultArray, filepathInfileIndexMap, err
		}
		// loggermdl.LogDebug("result", result)
		for _, val := range gjson.Parse(result).Array() {
			resultArray, _ = sjson.Set(resultArray, "-1", val.Value())
		}
		filepathInfileIndexMap[f] = updatedInfileIndexData
	}
	return resultArray, filepathInfileIndexMap, nil
}

func openFile(filePath string) (*os.File, error) {
	dir, _ := filepath.Split(filePath)
	if dir != "" {
		createError := filemdl.CreateDirectoryRecursive(dir)
		if errormdl.CheckErr(createError) != nil {
			return nil, errormdl.CheckErr(createError)
		}
	}
	return os.OpenFile(filePath, os.O_CREATE|os.O_RDWR, os.ModePerm)
}

func addFDBIndexEntry(index *Index, path string, rs *gjson.Result) (string, error) {
	json := `{}`
	rowID, err := GenRowID(path)
	if err != nil {
		loggermdl.LogError(err)
		return "", err
	}
	for _, indexField := range index.IndexFields {
		if rs.Get(indexField.Query).Value() == nil {
			return rowID, errormdl.Wrap("please provide value for index field: " + indexField.Query)
		}
		json, _ = sjson.Set(json, indexField.FieldName, rs.Get(indexField.Query).Value())
	}
	// if index.IsDynamicName {
	// 	rowID = rs.Get(index.IndexNameQuery).String()
	// 	if rowID == "" {
	// 		return rowID, errormdl.Wrap("please provide value for index field: " + index.IndexNameQuery)
	// 	}
	// } else {
	// 	rowID = index.IndexNameQuery
	// }
	json, _ = sjson.Set(json, "rowID", rowID)

	err = index.indexObj.Update(func(tx *buntdb.Tx) error {
		// optimization : if data not changed then dont update
		// path = strings.Trim(path, string(filepath.Separator))
		_, _, err := tx.Set(path, json, nil)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		return nil
	})
	return rowID, err
}

func updateFDBIndexEntry(index *Index, path string, updatedRs *gjson.Result) error {
	err := index.indexObj.Update(func(tx *buntdb.Tx) error {
		// optimization : if data not changed then dont update
		path = strings.Trim(path, string(filepath.Separator))
		_, _, err := tx.Set(path, updatedRs.String(), nil)
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		return nil
	})
	return err
}

// GetFDBIndexEntryByPath - returns value of specified key from index
func GetFDBIndexEntryByPath(index *Index, path string) (val string, err error) {
	path = filepath.Join(path)
	err = index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			if key == path {
				val = value
				return true
			}
			return true
		})
	})
	return
}

func addMigrationReplaceConfig(targetBasePath string, secParams securitymdl.FDBSecParams) error {
	configfilePath := filepath.Join(targetBasePath, MigrationConfigFilename)
	fp, err := openFile(configfilePath)
	defer func() {
		fp.Close()
	}()
	if err != nil {
		return err
	}
	migartionConfigStr, _ := sjson.Set("", MigrationTypeKeyword, MigrationTypeReplace)
	rs := gjson.Parse(migartionConfigStr)
	bucket := Bucket{}
	// err = saveDataInNormalBucketUsingFp(fp, rs)
	return saveDataInNormalBucketUsingFp(fp, &bucket, &rs, secParams)
}

func addMigrationUpdateConfig(targetBasePath string, secParams securitymdl.FDBSecParams) error {
	configfilePath := filepath.Join(targetBasePath, MigrationConfigFilename)
	fp, err := openFile(configfilePath)
	defer func() {
		fp.Close()
	}()
	if err != nil {
		return err
	}
	migartionConfigStr, _ := sjson.Set("", MigrationTypeKeyword, MigrationTypeUpdate)
	rs := gjson.Parse(migartionConfigStr)
	bucket := Bucket{}
	// err = saveDataInNormalBucketUsingFp(fp, rs)
	return saveDataInNormalBucketUsingFp(fp, &bucket, &rs, secParams)
}

// GetFDBIndexEntryMap - returns fdb index entries by fitering with queries
func GetFDBIndexEntryMap(index *Index, queries []string) (map[string]string, error) {
	entries := make(map[string]string, 0)
	err := index.indexObj.View(func(tx *buntdb.Tx) error {
		return tx.Ascend("", func(key, value string) bool {
			rsJSON := gjson.Parse("[" + value + "]")
			for i := 0; i < len(queries); i++ {
				rsJSON = rsJSON.Get(queries[i] + "#")
			}
			if rsJSON.Value() != nil && rsJSON.IsArray() && len(rsJSON.Array()) > 0 {
				entries[key] = value
			}
			return true
		})
	})
	return entries, err
}

// DataImporter is the interface that wraps the basic DataImport method.
type DataImporter interface {
	DataImport()
}

// DataExporter is the interface that wraps the basic DataExport method.
type DataExporter interface {
	DataExport()
}

// ZipImporter is a DataImporter
// allow to import fdb data from zip
type ZipImporter struct {
	FdbName    string
	IndexID    string
	SourcePath string
	Data       *gjson.Result
}

// ZipExporter is a DataExporter
// allow to export fdb data as zip
type ZipExporter struct {
	FdbName       string
	IndexID       string
	Queries       []string
	DestPath      string
	MigrationType string
}

// DataExport exports fdb data as zip
func (z ZipExporter) DataExport() (err error) {
	fdb, err := GetFDBInstance(z.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", z.FdbName)
		return errormdl.Wrap("fdb instance not found: " + z.FdbName)
	}
	index, ok := fdb.GetFDBIndex(z.IndexID)
	if !ok {
		return errormdl.Wrap("INDEX not found: " + z.IndexID)
	}
	sourcePath := ""
	timeStamp := time.Now().Nanosecond()
	targetBasePath, err := filepath.Abs(filepath.Join(filemdl.TempDir, strconv.Itoa(timeStamp)))
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	filteredKeyValMap, err := index.GetEntriesByQueries(z.Queries)
	if err != nil {
		return err
	}
	if len(filteredKeyValMap) == 0 {
		return errormdl.Wrap("no data found to export")
	}
	defer func() {
		// removes created zip
		filemdl.DeleteDirectory(targetBasePath)
	}()
	// copy data files
	for path := range filteredKeyValMap {
		sourcePath = filepath.Join(fdb.DBPath, path)
		targetPath := filepath.Join(targetBasePath, path)
		_, err := filemdl.CopyFile(sourcePath, targetPath, true)
		if err != nil {
			return err
		}
	}

	// copy index file
	// sourcePath = filepath.Join(fdb.DBPath, INDEXFOLDER, indexID)
	// _, err = filemdl.CopyFile(sourcePath, targetPath, true)
	targetPath := filepath.Join(targetBasePath, INDEXFOLDER, z.IndexID)
	err = AddIndexEntriesInFile(targetPath, filteredKeyValMap)
	if err != nil {
		return err
	}
	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}

	switch z.MigrationType {
	case MigrationTypeUpdate:
		err = addMigrationUpdateConfig(targetBasePath, secParams)
	case MigrationTypeReplace:
		err = addMigrationReplaceConfig(targetBasePath, secParams)
	default:
		return errormdl.Wrap("fail to export data: export operation not allowed on migration type - " + z.MigrationType)
	}

	if err != nil {
		loggermdl.LogError("fail to export data: ", err)
		return errormdl.Wrap("fail to export data: " + err.Error())
	}
	// make zip of copied data to destination folder
	// zip will have name of indexId
	destinationPath := filepath.Join(z.DestPath, z.IndexID)
	return filemdl.Zip(targetBasePath, destinationPath)
}

// DataImport imports data from zip
func (z ZipImporter) DataImport() (err error) {
	fdb, err := GetFDBInstance(z.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", z.FdbName)
		return errormdl.Wrap("fdb instance not found: " + z.FdbName)
	}
	index, ok := fdb.GetFDBIndex(z.IndexID)
	if !ok {
		loggermdl.LogError("index not found: ", z.IndexID)
		return errormdl.Wrap("index not found: " + z.IndexID)
	}
	archivePath := z.SourcePath
	if !filemdl.FileAvailabilityCheck(archivePath) {
		loggermdl.LogError("archive file not found at specified location: ", archivePath)
		return errormdl.Wrap("archive file not found at location: " + archivePath)
	}
	timeStamp := time.Now().Nanosecond()
	pathToExtractZip, err := filepath.Abs(filepath.Join(filemdl.TempDir, strconv.Itoa(timeStamp)))
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	err = filemdl.Unzip(archivePath, pathToExtractZip)
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	defer func() {
		// removes extracted files
		filemdl.DeleteDirectory(pathToExtractZip)
	}()
	childDirs, err := filemdl.ListDirectory(pathToExtractZip)
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	if len(childDirs) == 0 {
		loggermdl.LogError("no data found to import")
		return errormdl.Wrap("no data found to import")
	}
	if !childDirs[0].IsDir() {
		loggermdl.LogError("invalid archive file")
		return errormdl.Wrap("invalid archive file")
	}
	sourcePath := filepath.Join(pathToExtractZip, childDirs[0].Name())
	fdbBasePath := fdb.DBPath
	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}

	// loggermdl.LogDebug(sourcePath)
	migrationConfig, err := getMigrationConfig(sourcePath, secParams, z.Data)
	if err != nil {
		loggermdl.LogError("fail to get migration config", err)
		return errormdl.Wrap("invalid archived file")
	}
	migrationType := migrationConfig.Get(MigrationTypeKeyword).String()
	if migrationType != MigrationTypeReplace && migrationType != MigrationTypeUpdate {
		return errormdl.Wrap("import operation not allowed on migration type - " + migrationType)
	}
	indexFilePath := filepath.Join(fdb.DBPath, INDEXFOLDER, z.IndexID)
	err = filepath.Walk(sourcePath, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			loggermdl.LogError("err", err)
			return err
		}
		if info.IsDir() {
			return nil
		}
		//  ignore config file from copying
		if strings.Contains(path, MigrationConfigFilename) {
			return nil
		}

		foundAtIndex := strings.LastIndex(path, sourcePath)
		if foundAtIndex == -1 {
			return errormdl.Wrap("invalid archived file")
		}
		// loggermdl.LogDebug(path)

		// if migartion type is MigrationTypeUpdate then copy index entries from index files else replace index files
		if migrationType == MigrationTypeUpdate && strings.Contains(path, INDEXFOLDER) {
			// load index entries from
			err := ImportIndexEntries(path, fdb, z.IndexID)
			if err != nil {
				loggermdl.LogError("fail to load indexes from data", err)
				return errormdl.Wrap("fail to load indexes")
			}
			err = LogFDBIndexFile(indexFilePath, index)
			if err != nil {
				loggermdl.LogError("fail to add indexes: ", err)
				return errormdl.Wrap("fail to add indexes")
			}
			return nil
		}
		destPath := filepath.Join(fdbBasePath, path[foundAtIndex+len(sourcePath):])
		if !filemdl.FileAvailabilityCheck(destPath) {
			dir, _ := filepath.Split(destPath)
			err = filemdl.CreateDirectoryRecursive(dir)
			if err != nil {
				return err
			}
		}
		fileFpCache.Delete(destPath)
		return filemdl.AtomicReplaceFile(path, destPath)
	})
	if err != nil {
		loggermdl.LogError("fail to import data: ", err)
		return errormdl.Wrap("fail to import data: " + err.Error())
	}

	err = LoadFDBIndexFromFile(indexFilePath, fdb, z.IndexID)
	if err != nil {
		loggermdl.LogError("fail to add indexes", err)
		return errormdl.Wrap("fail to add indexes")
	}

	return nil
}

func getMigrationConfig(sourcePath string, secParams securitymdl.FDBSecParams, rs *gjson.Result) (*gjson.Result, error) {
	configPath := filepath.Join(sourcePath, MigrationConfigFilename)
	if !filemdl.FileAvailabilityCheck(configPath) {
		return nil, errormdl.Wrap("file not found")
	}
	fp, err := filemdl.Open(configPath)
	defer func() {
		fp.Close()
	}()
	dataByte, err := readNormalFileUsingFp(fp, rs, secParams)
	if err != nil {
		return nil, err
	}
	migrationConfig := gjson.ParseBytes(dataByte)
	return &migrationConfig, nil
}

// ImportIndexEntries -
func ImportIndexEntries(indexFilePath string, fdb *FDB, indexID string) error {
	index, found := fdb.GetFDBIndex(indexID)
	if !found {
		return errormdl.Wrap("index not found")
	}
	if !filemdl.FileAvailabilityCheck(indexFilePath) {
		return nil
	}
	fileData, err := filemdl.FastReadFile(indexFilePath)
	if err != nil {
		loggermdl.LogError("failed to load FDB index from: ", indexFilePath)
		return err
	}

	_, fileName := filepath.Split(indexFilePath)
	fileData, err = decryptData(fileData, fileName)
	if err != nil {
		loggermdl.LogError("failed to decrypt FDB index data: ", err)
		return errormdl.Wrap("failed to decrypt FDB index data: " + err.Error())
	}
	data := string(fileData)
	indexRecords := strings.Split(data, lineBreak)
	indexDataMap := make(map[string]string)
	for _, indexRecord := range indexRecords {
		indexValues := strings.Split(indexRecord, IndexKeyValSeperator)
		if len(indexValues) == 2 {
			indexDataMap[indexValues[0]] = indexValues[1]
		}
	}
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}

	// update index file by reading all data and updating index file
	return index.indexObj.Update(func(tx *buntdb.Tx) error {
		for key, val := range indexDataMap {
			_, _, err := tx.Set(key, val, nil)
			if err != nil {
				return err
			}
		}
		return nil
	})
}

// FileImporter is a DataImporter
// allow to import fdb data from exported folder
type FileImporter struct {
	FdbName    string
	IndexID    string
	SourcePath string
	Data       *gjson.Result
}

// FileExporter is a DataExporter
// allow to export fdb data in a folder
type FileExporter struct {
	FdbName       string
	IndexID       string
	Queries       []string
	DestPath      string
	MigrationType string
}

// DataExport exports fdb data in a folder
func (f FileExporter) DataExport() (err error) {
	fdb, err := GetFDBInstance(f.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", f.FdbName)
		return errormdl.Wrap("fdb instance not found: " + f.FdbName)
	}
	index, ok := fdb.GetFDBIndex(f.IndexID)
	if !ok {
		return errormdl.Wrap("INDEX not found: " + f.IndexID)
	}
	sourcePath := ""
	targetBasePath := filepath.Join(f.DestPath, f.IndexID)
	// loggermdl.LogDebug("targetBasePath", targetBasePath)

	filteredKeyValMap, err := GetFDBIndexEntryMap(index, f.Queries)
	if err != nil {
		return err
	}

	if len(filteredKeyValMap) == 0 {
		return errormdl.Wrap("no data found to export")
	}

	for path := range filteredKeyValMap {
		sourcePath = filepath.Join(fdb.DBPath, path)
		targetPath := filepath.Join(targetBasePath, path)
		_, err := filemdl.CopyFile(sourcePath, targetPath, true)
		if err != nil {
			return err
		}
	}

	// copy index file
	targetPath := filepath.Join(targetBasePath, INDEXFOLDER, f.IndexID)
	err = AddIndexEntriesInFile(targetPath, filteredKeyValMap)
	if err != nil {
		return err
	}

	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}

	switch f.MigrationType {
	case MigrationTypeUpdate:
		err = addMigrationUpdateConfig(targetBasePath, secParams)
	case MigrationTypeReplace:
		err = addMigrationReplaceConfig(targetBasePath, secParams)
	default:
		return errormdl.Wrap("export operation not allowed on migration type - " + f.MigrationType)
	}

	return err
}

// DataImport imports data from exported folder
func (f FileImporter) DataImport() (err error) {
	fdb, err := GetFDBInstance(f.FdbName)
	if err != nil {
		loggermdl.LogError("fdb instance not found: ", f.FdbName)
		return errormdl.Wrap("fdb instance not found: " + f.FdbName)
	}
	index, ok := fdb.GetFDBIndex(f.IndexID)
	if !ok {
		loggermdl.LogError("index not found: ", f.IndexID)
		return errormdl.Wrap("index not found: " + f.IndexID)
	}
	if !filemdl.FileAvailabilityCheck(f.SourcePath) {
		loggermdl.LogError("archive file not found at specified location: ", f.SourcePath)
		return errormdl.Wrap("archive file not found at location: " + f.SourcePath)
	}

	timeStamp := time.Now().Nanosecond()
	tempDir, err := filepath.Abs(filepath.Join(filemdl.TempDir, strconv.Itoa(timeStamp)))
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	err = filemdl.CopyDir(f.SourcePath, tempDir)
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("fail to copy data")
	}
	defer func() {
		filemdl.DeleteDirectory(tempDir)
	}()

	childDirs, err := filemdl.ListDirectory(tempDir)
	if err != nil {
		loggermdl.LogError("failed to import data: ", err)
		return errormdl.Wrap("invalid archived file")
	}
	if len(childDirs) == 0 {
		loggermdl.LogError("no data found to import")
		return errormdl.Wrap("no data found to import")
	}
	fdbBasePath := fdb.DBPath
	secParams := securitymdl.FDBSecParams{EnableSecurity: fdb.EnableSecurity, EnableCompression: fdb.EnableCompression}

	// loggermdl.LogDebug(f.SourcePath)
	migrationConfig, err := getMigrationConfig(tempDir, secParams, f.Data)
	if err != nil {
		loggermdl.LogError("fail to get migration config", err)
		return errormdl.Wrap("invalid archived file")
	}
	migrationType := migrationConfig.Get(MigrationTypeKeyword).String()
	if migrationType != MigrationTypeReplace && migrationType != MigrationTypeUpdate {
		return errormdl.Wrap("import operation not allowed on migration type - " + migrationType)
	}
	indexFilePath := filepath.Join(fdb.DBPath, INDEXFOLDER, f.IndexID)
	err = filepath.Walk(tempDir, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			loggermdl.LogError("err", err)
			return err
		}
		// loggermdl.LogError(path)
		if info.IsDir() {
			return nil
		}
		//  ignore config file from copying
		if strings.Contains(path, MigrationConfigFilename) {
			return nil
		}

		foundAtIndex := strings.LastIndex(path, tempDir)
		if foundAtIndex == -1 {
			return errormdl.Wrap("invalid archived file")
		}
		// if file is index file then copy index entries from index files
		if strings.Contains(path, INDEXFOLDER) {
			// load index entries from file
			err := ImportIndexEntries(path, fdb, f.IndexID)
			if err != nil {
				loggermdl.LogError("fail to import indexes", err)
				return errormdl.Wrap("fail to import indexes")
			}
			err = LogFDBIndexFile(indexFilePath, index)
			if err != nil {
				loggermdl.LogError("fail to import indexes: ", err)
				return errormdl.Wrap("fail to import indexes")
			}
			return nil
		}

		destPath := filepath.Join(fdbBasePath, path[foundAtIndex+len(tempDir):])
		if !filemdl.FileAvailabilityCheck(destPath) {
			dir, _ := filepath.Split(destPath)
			err = filemdl.CreateDirectoryRecursive(dir)
			if err != nil {
				return err
			}
		}
		// removing filepointer from memory
		fileFpCache.Delete(destPath)
		return filemdl.AtomicReplaceFile(path, destPath)
	})

	if err != nil {
		loggermdl.LogError("fail to import data: ", err)
		return errormdl.Wrap("fail to import data: " + err.Error())
	}

	return nil
}
