package corefdb

import (
	"os"
	"strings"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl/filepack"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/securitymdl"

	"github.com/tidwall/gjson"
	"github.com/tidwall/sjson"

	"path/filepath"
	"sync"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/dalmdl/lazywriter"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/cachemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/filemdl"
	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"

	"github.com/tidwall/buntdb"
)

const (
	// INDEXFOLDER -INDEXFOLDER
	INDEXFOLDER = "index"
	// LazyCallBackFnAppendBucket - LazyCallBackFnAppendBucket
	LazyCallBackFnAppendBucket = "LazyWriterAppendBucketCallBackFn"
	// LazyCallBackFnSaveIndex - LazyCallBackFnSaveIndex
	LazyCallBackFnSaveIndex = "LazyWriterCallBackFnAppendBucketSaveIndex"
)

var databases cachemdl.FastCacheHelper
var defaultDB string

var isLazyWriterEnabled = false
var defaultSecurityKey = []byte{}
var isSecurityEnabled = false
var fileFpCache Cache

func init() {
	databases.Setup(1, 1000, 1000)
	fpCache, _ := NewCache()
	fileFpCache = *fpCache
}

// Init - initializes bundbmdl
func Init(isSecurityRequired, isCompressionRequired, isLazyWriterEnable bool, securityKey string) {
	filepack.Init(isSecurityRequired, isCompressionRequired, securityKey)
	defaultSecurityKey = []byte(securityKey)
	isLazyWriterEnabled = isLazyWriterEnable
	isSecurityEnabled = isSecurityRequired
}

// FDB - FDB
type FDB struct {
	DBName              string
	DBPath              string `json:"dbPath"`
	EnableSecurity      bool   `json:"enableSec"` // if enabled, fdb files will be encrypted
	EnableCompression   bool   `json:"enableCmp"` // if enabled, fdb files will be compressed and then encrypted
	indices             map[string]*Index
	indexMux            sync.Mutex
	buckets             map[string]*Bucket
	bLocker             sync.Mutex
	restoreFileFromPack bool
}

// CreateFDBInstance - creates fdb instance
func CreateFDBInstance(dbPath, dbName string, isDefault bool) (*FDB, error) {
	fdb := &FDB{
		DBPath:              dbPath,
		indices:             make(map[string]*Index),
		indexMux:            sync.Mutex{},
		buckets:             make(map[string]*Bucket),
		bLocker:             sync.Mutex{},
		restoreFileFromPack: true,
		DBName:              dbName,
	}
	if isDefault {
		defaultDB = dbName
	}
	databases.SetNoExpiration(dbName, fdb)
	return fdb, nil
}

// EnableFDBSecurity enables security. Files will be encrypted.
func (fdb *FDB) EnableFDBSecurity(sec bool) {
	if !sec {
		return
	}
	fdb.EnableSecurity = sec
}

// EnableFDBCompression enables security. Files will be encrypted.
func (fdb *FDB) EnableFDBCompression(cmp bool) {
	if !cmp {
		return
	}
	fdb.EnableCompression = cmp
}

// GetBucketByName - reaturn bucket with specified bucket name
func (f *FDB) GetBucketByName(bucketName string) *Bucket {
	for _, val := range f.buckets {
		if val.BucketNameQuery == bucketName {
			return val
		}
	}
	return nil
}

// SetFileRestoreFormPackFlag - set whether to restore file from pack while performing Get operation
func (f *FDB) SetFileRestoreFormPackFlag(restoreFile bool) {
	f.restoreFileFromPack = restoreFile
}

// GetFDBInstance - returns fdb instance
func GetFDBInstance(dbName string) (*FDB, error) {
	if dbName == "" {
		dbName = defaultDB
	}

	rawDB, ok := databases.Get(dbName)
	if !ok {
		loggermdl.LogError("Database instance not found")
		return nil, errormdl.Wrap("Database instance not found")
	}
	fdb, ok := rawDB.(*FDB)
	if !ok {
		loggermdl.LogError("Can not cast object into *FDB")
		return nil, errormdl.Wrap("Can not cast object into *FDB")
	}
	return fdb, nil
}

// GetFDBIndex - returns index
func (f *FDB) GetFDBIndex(indexName string) (*Index, bool) {
	index, ok := f.indices[indexName]
	return index, ok
}

// CreateIndex - Creates index
func (f *FDB) CreateIndex(index *Index) error {
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	err := index.indexObj.CreateIndex(index.IndexID, "*", fns...)
	if err != nil {
		return err
	}
	return nil
}

// GetBucketIndexes ; returns list of indexes of bucket
func (f *FDB) GetBucketIndexes(bucketName string) ([]*Index, error) {

	var bucket *Bucket
	var foundBucket = false
	for _, bucketObj := range f.buckets {
		if bucketObj.BucketNameQuery == bucketName {
			foundBucket = true
			bucket = bucketObj
			break
		}

	}

	if !foundBucket {
		return nil, errormdl.Wrap("Bucket not found")
	}

	indexList := make([]*Index, 0)
	for _, indexID := range bucket.Indices {
		index, ok := f.indices[indexID]

		if !ok {
			return nil, errormdl.Wrap("index not found")
		}
		indexList = append(indexList, index)
	}
	return indexList, nil
}

// ReIndex - performs reindexing
func (f *FDB) ReIndex(indexID string) error {

	index, found := f.GetFDBIndex(indexID)
	if !found {
		return errormdl.Wrap("index not found")
	}
	// find path to start file walk
	pathToStartWalk := f.DBPath
	for _, bucketID := range index.BucketSequence {
		bucket := f.buckets[bucketID]
		if bucket.IsDynamicName {
			break
		}
		pathToStartWalk = filepath.Join(pathToStartWalk, bucket.BucketNameQuery)
	}

	// get required data for index file
	indexDataMap, err := getFDBIndexData(pathToStartWalk, index, f.DBPath)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// create or replace index
	var fns []func(a, b string) bool
	for _, idx := range index.IndexFields {
		fns = append(fns, buntdb.IndexJSON(idx.FieldName))
	}
	index.indexObj.Close()
	indexFilePath := filepath.Join(f.DBPath, INDEXFOLDER, index.IndexID)
	if filemdl.FileAvailabilityCheck(indexFilePath) {
		err = filemdl.DeleteFile(indexFilePath)
		if err != nil {
			return err
		}
	}
	db, err := f.GetIndexDB(index)
	if err != nil {
		return err
	}
	index.indexObj = db
	err = index.indexObj.ReplaceIndex(index.IndexID, "*", fns...)
	if err != nil {
		loggermdl.LogError(err)
		return err
	}
	// update index file by reading all data and updating index file
	return index.indexObj.Update(func(tx *buntdb.Tx) error {

		for indexKey, indexData := range indexDataMap {
			_, _, err := tx.Set(indexKey, indexData, nil)
			if err != nil {
				loggermdl.LogError(err)
				return err
			}
		}

		return nil
	})
}

// getFDBIndexData - returns index data with index fields recursively from specified direcory
func getFDBIndexData(path string, index *Index, dbPath string) (map[string]string, error) {

	indexDataMap := make(map[string]string)
	if !filemdl.FileAvailabilityCheck(path) {
		return nil, errormdl.Wrap("invalid path: " + path)
	}

	dbPath = filepath.Join(dbPath) + string(filepath.Separator)
	ferr := filemdl.Walk(path, func(filePath string, info os.FileInfo, err error) error {
		if err != nil {
			loggermdl.LogError(err)
			return nil
		}

		if !info.IsDir() {
			fileData, err := filemdl.ReadFile(filePath)
			if err != nil {
				loggermdl.LogError(err)
				return nil
			}

			dataObj := gjson.Parse(string(fileData))
			indexDataObj := ""
			for _, indexField := range index.IndexFields {
				indexDataObj, _ = sjson.Set(indexDataObj, indexField.FieldName, dataObj.Get(indexField.Query).String())
			}

			if indexDataObj != "" {
				pathToSave := strings.TrimPrefix(filePath, dbPath)
				indexDataMap[pathToSave] = indexDataObj
			}
		}

		return nil
	})
	if ferr != nil {
		return indexDataMap, ferr
	}
	return indexDataMap, nil
}

// GetLazyCallBackFunc - return callback functions for lazywriter
func GetLazyCallBackFunc(funcName string) (lazywriter.SaveDataFn, error) {
	if funcName == LazyCallBackFnAppendBucket {
		return lazyCallBackFnAppendBucket, nil
	} else if funcName == LazyCallBackFnSaveIndex {
		return lazyCallBackFnSaveIndex, nil
	} else {
		return nil, errormdl.Wrap("func not found for: " + funcName)
	}
}

var lazyCallBackFnAppendBucket lazywriter.SaveDataFn = func(bucketId string, data *lazywriter.LazyCacheObject) {
	dataInLazyMemory, ok := data.InterfaceData.(string)
	if !ok {
		return
	}
	_, _, err := filemdl.AppendDataToFile(data.FileName, []byte(dataInLazyMemory+"\r\n"), true)
	if err != nil {
		loggermdl.LogError(err)
		return
	}
	data.InterfaceData = nil
	err = AppendMaster.ClearLazyObjInterfaceData(data.Identifier)
	if err != nil {
		loggermdl.LogError(err)
		return
	}
	AppendLazyObjHolder.Set(data.Identifier, *data)
}

var lazyCallBackFnSaveIndex lazywriter.SaveDataFn = func(indexID string, data *lazywriter.LazyCacheObject) {
	indexData, ok := data.InterfaceData.(*Index)
	if !ok {
		return
	}

	err := LogFDBIndexFile(data.FileName, indexData)
	if err != nil {
		loggermdl.LogError(err)
		return
	}
}

// GenRowID generates fash for the given filename. The length of hash is 16
func GenRowID(name string) (string, error) {
	name = strings.ReplaceAll(filepath.Clean(name), string(filepath.Separator), "")
	rowID, err := securitymdl.GetHash(name)
	if err != nil {
		return "", err
	}

	if len(rowID) > 16 {
		return rowID[:16], nil
	}

	return rowID, nil
}
