package corefdb

import (
	"os"
	"sync"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/errormdl"

	"github.com/dgraph-io/ristretto"
	"github.com/tidwall/gjson"
)

var (
	// CacheConfig - used for creating new cache instance
	CacheConfig = ristretto.Config{
		MaxCost:     1000,
		NumCounters: 10000,
		BufferItems: 64,
		OnEvict:     onFilepointerEvict,
	}
)

// File -
type File struct {
	file        *os.File
	lock        *sync.Mutex
	InfileIndex *gjson.Result
}

// Cache -
type Cache struct {
	cache ristretto.Cache
	cfg   *ristretto.Config
}

// Get -
func (c Cache) Get(key string) (*File, error) {
	val, found := c.cache.Get(key)
	if !found {
		return nil, errormdl.Wrap("not found")
	}
	fileInstance, ok := val.(*File)
	if !ok {
		return nil, errormdl.Wrap("invalid cache val")
	}
	return fileInstance, nil
}

// Set - sets filepointer and infile index data in cache
func (c Cache) Set(fp *os.File, infileIndexData *gjson.Result) *File {
	rfile := File{
		file:        fp,
		InfileIndex: infileIndexData,
		lock:        &sync.Mutex{},
	}
	file, err := c.Get(fp.Name())
	if err == nil {
		rfile.lock = file.lock
	}
	c.cache.Set(fp.Name(), &rfile, 1)
	return &rfile
}

// Delete -
func (c Cache) Delete(key string) {
	f, err := c.Get(key)
	if err != nil {
		return
	}
	f.file.Close()
	c.cache.Del(key)
}

// DeleteAll -
func (c Cache) DeleteAll() {
	c.cache.Clear()
}

// NewCache - creates new Cache
func NewCache() (*Cache, error) {
	c, err := ristretto.NewCache(&CacheConfig)
	if err != nil {
		return nil, err
	}

	cacheInstance := Cache{
		cache: *c,
		cfg:   &CacheConfig,
	}
	return &cacheInstance, nil
}

func onFilepointerEvict(hashes, hash2 uint64, val interface{}, cost int64) {
	fileInstance, ok := val.(*File)
	if !ok {
		return
	}
	fileInstance.lock.Lock()
	defer func() {
		fileInstance.lock.Unlock()
	}()
	fileInstance.file.Close()
	// loggermdl.LogError("evicted - ", fileInstance.file.Name())
}
