package dgraph

import (
	"context"
	"errors"

	"github.com/tidwall/sjson"
	"google.golang.org/grpc"

	"corelab.mkcl.org/MKCLOS/coredevelopmentplatform/corepkgv2/loggermdl"
	"github.com/dgraph-io/dgo"
	"github.com/dgraph-io/dgo/protos/api"
	"github.com/tidwall/gjson"
)

type DGraph interface {
	ConnInit()
	Create()
	Drop()
	Get()
	Mutate()
}

/** ------------------------------------------ **/
//
// DGraphHost Parameters
//
// ==========================================
//
//* Server = Database Host IP Address
//
//* Port = Database Port
//
//* Username = Database Username
//
//* Password = Database Password
//
//* MaxConnectionPools = Connection pool to store. Default is 10.
type DGraphHost struct {
	Server             string `json:"server"`
	Port               string `json:"port"`
	Username           string `json:"username"`
	Password           string `json:"password"`
	MaxConnectionPools int    `json:"maxConnPools"`
}

/** ------------------------------------------ **/
//
// DGraphObj Parameters
//
// ==========================================
//
//* Dbg = Dgraph Connection Object
//
//* Ctx = Background Context
//
//* Txn = Transaction Object
type DGraphObj struct {
	Dbg *dgo.Dgraph
	Ctx context.Context
	Txn *dgo.Txn
}

/** ------------------------------------------ **/
//
// Mutate Operation Parameters
//
// ==========================================
//
//* Dbg = Dgraph Connection Object
//
//* Ctx = Background Context
//
//* Txn = Transaction Object
//
//* EnableTxn = Hold transaction for multiple write ops and commit once everything is done
//
//* Query = Data object
type MutateOperation struct {
	EnableTxn bool
	JSONData  string
}

/** ------------------------------------------ **/
//
// DROP Operation Parameters
//
// ==========================================
//
//* Operation = 1 || 2 || 3 || 4
//
//** "schema" = Drop complete Schema with data
//
//** "data" = Drop complete data but maintain the schema
//
//** "attr" = Drop specific Attribute completely from data
//
//* AttributeName = REQUIRED IF OPERATION TYPE = 3
type DropParams struct {
	Operation     string
	AttributeName string
}

// Drop Operation Method
func (d *DGraphObj) Drop(q *DropParams) error {
	switch q.Operation {
	// Drop complete Schema with data
	case "schema":
		err := d.Dbg.Alter(d.Ctx, &api.Operation{DropOp: api.Operation_ALL})
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		break

	// Drop complete data but maintain the schema
	case "data":
		err := d.Dbg.Alter(d.Ctx, &api.Operation{DropOp: api.Operation_DATA})
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		break

	// Drop specific Attribute completely from data
	case "attr":
		err := d.Dbg.Alter(d.Ctx, &api.Operation{DropOp: api.Operation_ATTR, DropValue: q.AttributeName})
		if err != nil {
			loggermdl.LogError(err)
			return err
		}
		break

	default:
		return errors.New("Invalid drop operation: " + q.Operation)
	}

	return nil
}

// Connect to DB via Pools
// Return connection object
func (d *DGraphHost) ConnInit() (*dgo.Dgraph, error) {
	serverAddr := d.Server + ":" + d.Port
	conn, err := grpc.Dial(serverAddr, grpc.WithInsecure())
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}

	return dgo.NewDgraphClient(
		api.NewDgraphClient(conn),
	), nil
}

// Create new schema in database with input DQuery json
func (q *DGraphObj) Create(data string) error {
	err := q.Dbg.Alter(q.Ctx, &api.Operation{
		Schema: data,
	})

	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	return nil
}

// Execute Query with DB Obj, Txn Obj and Query string
// Returns gjson result string
func (q *DGraphObj) Get(query string) (*gjson.Result, error) {

	txn := q.Dbg.NewReadOnlyTxn().BestEffort()
	defer txn.Discard(q.Ctx)

	resp, err := txn.Query(q.Ctx, query)
	if err != nil {
		loggermdl.LogError(err)
		return nil, err
	}

	rs := gjson.ParseBytes(resp.Json)
	return &rs, nil
}

// Mutates query to database
// Returns error
func (d *DGraphObj) Mutate(m *MutateOperation) error {

	commitNow := false

	// If transaction object input is nil, return error
	if m.EnableTxn && d.Txn == nil {
		loggermdl.LogError("Transaction object is nil")
		return errors.New("Transaction object is nil")
	}

	// If commit is True, discard the transaction immediately after commit
	if !m.EnableTxn {
		d.Txn = d.Dbg.NewTxn()
		commitNow = true // Commit the transaction immediately
		defer d.Txn.Discard(d.Ctx)
	}

	// Parse gjson and set the query
	res := gjson.Parse(m.JSONData)
	obj, err := sjson.Set("", "set", res.Value())
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	_, err = d.Txn.Mutate(d.Ctx, &api.Mutation{SetJson: []byte(obj), CommitNow: commitNow})
	if err != nil {
		loggermdl.LogError(err)
		return err
	}

	return nil
}
